import pytest


@pytest.fixture(scope='module')
def docker_registry():
    from docker import DockerClient
    from pachypy.registry import DockerRegistryAdapter
    return DockerRegistryAdapter(DockerClient())


def test_docker_get_image_digest(docker_registry):
    from pachypy.registry import RegistryException
    digest = docker_registry.get_image_digest('alpine:latest')
    assert digest.startswith('sha256:')
    with pytest.raises(RegistryException):
        docker_registry.get_image_digest('repo_not_existing:latest')
    with pytest.raises(RegistryException):
        docker_registry.get_image_digest('alpine:tag_not_existing')


def test_split_image_string():
    from pachypy.registry import AmazonECRAdapter
    assert AmazonECRAdapter._split_image_string('a.b/c:d@sha1:e') == ('a.b', 'c', 'd')
    assert AmazonECRAdapter._split_image_string('a.b/c:d') == ('a.b', 'c', 'd')
    assert AmazonECRAdapter._split_image_string('a.b/c') == ('a.b', 'c', None)
    assert AmazonECRAdapter._split_image_string('a.b/c/d') == ('a.b', 'c/d', None)
    assert AmazonECRAdapter._split_image_string('a.b/c/d:e') == ('a.b', 'c/d', 'e')
    assert AmazonECRAdapter._split_image_string('a/b') == ('docker.io', 'a/b', None)
    assert AmazonECRAdapter._split_image_string('a/b:c') == ('docker.io', 'a/b', 'c')
    assert AmazonECRAdapter._split_image_string('a:b') == ('docker.io', 'a', 'b')
    with pytest.raises(ValueError):
        AmazonECRAdapter._split_image_string('a:b:c@d')



import os
import time
import json
from typing import List, Dict, Generator, Callable, Union, Optional, TypeVar, Any, cast

import grpc
import pandas as pd
from google.protobuf.json_format import MessageToJson
from google.protobuf.timestamp_pb2 import Timestamp
from python_pachyderm.client.pfs.pfs_pb2 import (
    Repo, Commit, Branch, File,
    ListRepoRequest, ListCommitRequest, ListBranchRequest, GlobFileRequest, GetFileRequest,
    CreateRepoRequest, DeleteRepoRequest,
    CreateBranchRequest, DeleteBranchRequest,
    StartCommitRequest, FinishCommitRequest, DeleteCommitRequest, FlushCommitRequest,
    PutFileRequest, DeleteFileRequest,
    RESERVED as FILETYPE_RESERVED, FILE as FILETYPE_FILE, DIR as FILETYPE_DIR,
    NONE as DELIMITER_NONE, JSON as DELIMITER_JSON, LINE as DELIMITER_LINE, SQL as DELIMITER_SQL, CSV as DELIMITER_CSV
)
from python_pachyderm.client.pfs.pfs_pb2_grpc import APIStub as PfsAPIStub
from python_pachyderm.client.pps.pps_pb2 import (
    Pipeline, Job, Input, Transform, Datum,
    ListPipelineRequest, ListJobRequest, ListDatumRequest, GetLogsRequest, DeleteJobRequest,
    CreatePipelineRequest, DeletePipelineRequest, StartPipelineRequest, StopPipelineRequest,
    InspectPipelineRequest, InspectJobRequest, InspectDatumRequest,
    FAILED as DATUM_FAILED, SUCCESS as DATUM_SUCCESS, SKIPPED as DATUM_SKIPPED, STARTING as DATUM_STARTING,
    JOB_STARTING, JOB_RUNNING, JOB_MERGING, JOB_SUCCESS, JOB_FAILURE, JOB_KILLED,
    PIPELINE_STARTING, PIPELINE_RUNNING, PIPELINE_RESTARTING, PIPELINE_FAILURE, PIPELINE_PAUSED, PIPELINE_STANDBY,
)
from python_pachyderm.client.pps.pps_pb2_grpc import APIStub as PpsAPIStub
from python_pachyderm.client.version.versionpb.version_pb2 import Version
from python_pachyderm.client.version.versionpb.version_pb2_grpc import APIStub as VersionAPIStub

from .utils import invert_dict, to_timestamp, to_timedelta


T = TypeVar('T')

pipeline_state_mapping = {
    PIPELINE_STARTING: 'starting',
    PIPELINE_RUNNING: 'running',
    PIPELINE_RESTARTING: 'restarting',
    PIPELINE_FAILURE: 'failure',
    PIPELINE_PAUSED: 'paused',
    PIPELINE_STANDBY: 'standby',
    'PIPELINE_STARTING': 'starting',
    'PIPELINE_RUNNING': 'running',
    'PIPELINE_RESTARTING': 'restarting',
    'PIPELINE_FAILURE': 'failure',
    'PIPELINE_PAUSED': 'paused',
    'PIPELINE_STANDBY': 'standby',
}

job_state_mapping = {
    JOB_STARTING: 'starting',
    JOB_RUNNING: 'running',
    JOB_MERGING: 'merging',
    JOB_FAILURE: 'failure',
    JOB_SUCCESS: 'success',
    JOB_KILLED: 'killed',
    'JOB_STARTING': 'starting',
    'JOB_RUNNING': 'running',
    'JOB_MERGING': 'merging',
    'JOB_FAILURE': 'failure',
    'JOB_SUCCESS': 'success',
    'JOB_KILLED': 'killed',
}

datum_state_mapping = {
    DATUM_FAILED: 'failed',
    DATUM_SUCCESS: 'success',
    DATUM_SKIPPED: 'skipped',
    DATUM_STARTING: 'starting',
    'FAILED': 'failed',
    'SUCCESS': 'success',
    'SKIPPED': 'skipped',
    'STARTING': 'starting',
}

file_type_mapping = {
    FILETYPE_RESERVED: 'reserved',
    FILETYPE_FILE: 'file',
    FILETYPE_DIR: 'dir',
    'RESERVED': 'reserved',
    'FILE': 'file',
    'DIR': 'dir',
}


class PachydermError(Exception):

    def __init__(self, message: str, code=None):
        super().__init__(message)
        try:
            self.status_code = code.value[0]
            self.status = code.value[1]
        except (AttributeError, KeyError):
            self.status_code = None
            self.status = None


def retry(f: T) -> T:
    def retry_wrapper(self, *args, **kwargs):
        adapter = self.adapter if hasattr(self, 'adapter') else self
        try:
            return f(self, *args, **kwargs)
        except grpc._channel._Rendezvous as e:
            if e.code().value[1] == 'unavailable' and adapter._retries < adapter._max_retries:
                if adapter.check_connectivity():
                    adapter._retries += 1
                    return retry_wrapper(self, *args, **kwargs)
            raise PachydermError(e.details(), e.code())
        else:
            adapter._retries = 0
            adapter._connectable = True
    return cast(T, retry_wrapper)


class PachydermAdapter:

    """Adapter class handling communication with Pachyderm.

    It is effectively a wrapper around the python_pachyderm package.
    """

    def __init__(self, host: Optional[str] = None, port: Optional[int] = None):
        if host is None:
            host = os.getenv('PACHD_ADDRESS')
        if host is None:
            try:
                with open(os.path.expanduser('~/.pachyderm/config.json'), 'r') as f:
                    config = json.load(f)
                    host = config['v1']['pachd_address']
            except (json.JSONDecodeError, KeyError):
                pass
        if host is not None and ':' in host:
            host_split = host.split(':')
            host = host_split[0]
            port = int(host_split[1])

        self.host = host or 'localhost'
        self.port = port or 30650
        self.channel = grpc.insecure_channel(f'{self.host}:{self.port}')
        self.pfs_stub = PfsAPIStub(self.channel)
        self.pps_stub = PpsAPIStub(self.channel)
        self.version_stub = VersionAPIStub(self.channel)

        self._retries = 0
        self._max_retries = 1
        self._connectable: Optional[bool] = None

    def check_connectivity(self, timeout: float = 10.0) -> bool:
        """Checks the connectivity to pachd. Tries to connect if not currently connected.

        The gRPC channel connectivity knows 5 states:
        0 = idle, 1 = connecting, 2 = ready, 3 = transient failure, 4 = shutdown.

        Args:
            timeout: Timeout in seconds.

        Returns:
            True if the connectivity state is ready (2), False otherwise.
        """
        connectivity = 0
        timeout = time.time() + timeout
        connectivity = self.channel._channel.check_connectivity_state(True)
        while connectivity < 2:
            if time.time() > timeout:
                connectivity = 5
                break
            time.sleep(0.001)
            connectivity = self.channel._channel.check_connectivity_state(False)
        self._connectable = connectivity == 2
        return self._connectable

    @retry
    def list_repos(self) -> pd.DataFrame:
        res = []
        for repo in self.pfs_stub.ListRepo(ListRepoRequest()).repo_info:
            res.append({
                'repo': repo.repo.name,
                'size_bytes': repo.size_bytes,
                'branches': [b.name for b in repo.branches],
                'created': to_timestamp(repo.created.seconds, repo.created.nanos),
            })
        return pd.DataFrame(res, columns=['repo', 'size_bytes', 'branches', 'created']) \
            .astype({'size_bytes': 'int', 'created': 'datetime64[ns, UTC]'})

    @retry
    def list_repo_names(self) -> List[str]:
        repos = self.pfs_stub.ListRepo(ListRepoRequest()).repo_info
        return [r.repo.name for r in repos]

    @retry
    def list_branch_heads(self, repo: str) -> Dict[str, str]:
        branches = self.pfs_stub.ListBranch(ListBranchRequest(repo=Repo(name=repo))).branch_info
        return {b.branch.name: b.head.id for b in branches}

    @retry
    def list_commits(self, repo: str, n: int = 20) -> pd.DataFrame:
        i = 1
        res = []
        commit_branches = invert_dict(self.list_branch_heads(repo))
        stream = self.pfs_stub.ListCommitStream(ListCommitRequest(repo=Repo(name=repo)))
        for commit in stream:
            res.append({
                'repo': commit.commit.repo.name,
                'commit': commit.commit.id,
                'parent_commit': commit.parent_commit.id,
                'branches': commit_branches.get(commit.commit.id, []),
                'size_bytes': commit.size_bytes,
                'started': to_timestamp(commit.started.seconds, commit.started.nanos),
                'finished': to_timestamp(commit.finished.seconds, commit.finished.nanos),
            })
            i += 1
            if n is not None and i > n:
                stream.cancel()
                break
        return pd.DataFrame(res, columns=['repo', 'commit', 'branches', 'size_bytes', 'started', 'finished', 'parent_commit']) \
            .astype({'size_bytes': 'int', 'started': 'datetime64[ns, UTC]', 'finished': 'datetime64[ns, UTC]'})

    @retry
    def list_files(self, repo: str, branch: Optional[str] = 'master', commit: Optional[str] = None, glob: str = '**') -> pd.DataFrame:
        if branch is None and commit is None:
            raise ValueError('branch and commit cannot both be None')
        res = []
        branch_heads = self.list_branch_heads(repo)
        commit_branches = invert_dict(branch_heads)
        if commit is None and branch is not None:
            commit = branch_heads.get(branch)
        if commit is not None:
            commit = Commit(repo=Repo(name=repo), id=commit)
            for file in self.pfs_stub.GlobFileStream(GlobFileRequest(commit=commit, pattern=glob)):
                res.append({
                    'repo': file.file.commit.repo.name,
                    'commit': file.file.commit.id,
                    'branches': commit_branches.get(file.file.commit.id, []),
                    'path': file.file.path,
                    'type': file_type_mapping.get(file.file_type, 'unknown'),
                    'size_bytes': file.size_bytes,
                    'committed': to_timestamp(file.committed.seconds, file.committed.nanos),
                })
        return pd.DataFrame(res, columns=['repo', 'commit', 'branches', 'path', 'type', 'size_bytes', 'committed']) \
            .astype({'size_bytes': 'int', 'committed': 'datetime64[ns, UTC]'})

    @retry
    def list_pipelines(self) -> pd.DataFrame:
        def input_string(i: Input) -> str:
            if i.cross:
                return '(' + ' ⨯ '.join([input_string(j) for j in i.cross]) + ')'
            elif i.union:
                return '(' + ' ∪ '.join([input_string(j) for j in i.union]) + ')'
            elif i.atom.name:
                name = i.atom.name + ('/' + i.atom.branch if i.atom.branch != 'master' else '')
                return str(name + ':' + i.atom.glob)
            elif i.pfs.name:
                name = i.pfs.name + ('/' + i.pfs.branch if i.pfs.branch != 'master' else '')
                return str(name + ':' + i.pfs.glob)
            elif i.cron.name:
                return str(i.cron.name)
            elif i.git.name:
                return str(i.git.name + ('/' + i.git.branch if i.git.branch != 'master' else ''))
            else:
                return '?'

        def input_repos(i: Input) -> Generator[str, None, None]:
            cross_or_union = i.cross or i.union
            if cross_or_union:
                for j in cross_or_union:
                    yield from input_repos(j)
            elif i.atom.repo:
                yield i.atom.repo
            elif i.pfs.repo:
                yield i.pfs.repo

        res = []
        pipelines = self.pps_stub.ListPipeline(ListPipelineRequest()).pipeline_info
        for pipeline in pipelines:
            res.append({
                'pipeline': pipeline.pipeline.name,
                'image': pipeline.transform.image,
                'cron_spec': ', '.join([cron['spec'] for cron in self._get_pipeline_input_cron_specs(pipeline.input)]),
                'input': input_string(pipeline.input),
                'input_repos': list(input_repos(pipeline.input)),
                'output_branch': pipeline.output_branch,
                'parallelism_constant': pipeline.parallelism_spec.constant,
                'parallelism_coefficient': pipeline.parallelism_spec.coefficient,
                'datum_tries': pipeline.datum_tries,
                'max_queue_size': pipeline.max_queue_size,
                'jobs_running': pipeline.job_counts[JOB_RUNNING],
                'jobs_success': pipeline.job_counts[JOB_SUCCESS],
                'jobs_failure': pipeline.job_counts[JOB_FAILURE],
                'created': to_timestamp(pipeline.created_at.seconds, pipeline.created_at.nanos),
                'state': pipeline_state_mapping.get(pipeline.state, 'unknown'),
            })
        return pd.DataFrame(res, columns=[
            'pipeline', 'state', 'image', 'cron_spec', 'input', 'input_repos', 'output_branch',
            'parallelism_constant', 'parallelism_coefficient', 'datum_tries', 'max_queue_size',
            'jobs_running', 'jobs_success', 'jobs_failure', 'created',
        ]).astype({
            'parallelism_constant': 'int',
            'parallelism_coefficient': 'float',
            'datum_tries': 'int',
            'jobs_running': 'int',
            'jobs_success': 'int',
            'jobs_failure': 'int',
            'created': 'datetime64[ns, UTC]',
        })

    @retry
    def list_pipeline_names(self) -> List[str]:
        pipelines = self.pps_stub.ListPipeline(ListPipelineRequest()).pipeline_info
        return [p.pipeline.name for p in pipelines]

    @retry
    def list_jobs(self, pipeline: Optional[str] = None, n: int = 20) -> pd.DataFrame:
        i = 1
        res = []
        jobs_stream = self.pps_stub.ListJobStream(ListJobRequest(pipeline=Pipeline(name=pipeline)))
        for job in jobs_stream:
            res.append({
                'job': job.job.id,
                'pipeline': job.pipeline.name,
                'state': job_state_mapping.get(job.state, 'unknown'),
                'started': to_timestamp(job.started.seconds, job.started.nanos),
                'finished': to_timestamp(job.finished.seconds, job.finished.nanos),
                'restart': job.restart,
                'data_processed': job.data_processed,
                'data_skipped': job.data_skipped,
                'data_total': job.data_total,
                'download_time': to_timedelta(job.stats.download_time.seconds, job.stats.download_time.nanos),
                'process_time': to_timedelta(job.stats.process_time.seconds, job.stats.process_time.nanos),
                'upload_time': to_timedelta(job.stats.upload_time.seconds, job.stats.upload_time.nanos),
                'download_bytes': job.stats.download_bytes,
                'upload_bytes': job.stats.upload_bytes,
                'output_commit': job.output_commit.id,
            })
            i += 1
            if n is not None and i > n:
                jobs_stream.cancel()
                break
        return pd.DataFrame(res, columns=[
            'job', 'pipeline', 'state', 'started', 'finished', 'restart',
            'data_processed', 'data_skipped', 'data_total',
            'download_time', 'process_time', 'upload_time',
            'download_bytes', 'upload_bytes', 'output_commit'
        ]).astype({
            'started': 'datetime64[ns, UTC]',
            'finished': 'datetime64[ns, UTC]',
            'restart': 'int',
            'data_processed': 'int',
            'data_skipped': 'int',
            'data_total': 'int',
            'download_time': 'timedelta64[ns]',
            'process_time': 'timedelta64[ns]',
            'upload_time': 'timedelta64[ns]',
            'download_bytes': 'float',
            'upload_bytes': 'float',
        })

    @retry
    def list_datums(self, job: str) -> pd.DataFrame:
        file_type_mapping = {
            FILETYPE_RESERVED: 'reserved',
            FILETYPE_FILE: 'file',
            FILETYPE_DIR: 'dir',
        }
        res = []
        datums_stream = self.pps_stub.ListDatumStream(ListDatumRequest(job=Job(id=job)))
        for datum in datums_stream:
            for data in datum.datum_info.data:
                res.append({
                    'job': datum.datum_info.datum.job.id,
                    'datum': datum.datum_info.datum.id,
                    'state': datum_state_mapping.get(datum.datum_info.state, 'unknown'),
                    'repo': data.file.commit.repo.name,
                    'commit': data.file.commit.id,
                    'path': data.file.path,
                    'type': file_type_mapping.get(data.file_type, 'unknown'),
                    'size_bytes': data.size_bytes,
                    'committed': to_timestamp(data.committed.seconds, data.committed.nanos),
                })
        return pd.DataFrame(res, columns=['job', 'datum', 'state', 'repo', 'path', 'type', 'size_bytes', 'commit', 'committed']) \
            .astype({'size_bytes': 'int', 'committed': 'datetime64[ns, UTC]'})

    @retry
    def get_logs(self, pipeline: Optional[str] = None, job: Optional[str] = None, master: bool = False) -> pd.DataFrame:
        pipeline = Pipeline(name=pipeline) if pipeline else None
        job = Job(id=job) if job else None
        if pipeline is None and job is None:
            raise ValueError('One of `pipeline` or `job` must be specified')
        res = []
        logs = self.pps_stub.GetLogs(GetLogsRequest(pipeline=pipeline, job=job, data_filters=tuple(), master=master))
        for msg in logs:
            message = msg.message.strip()
            if message:
                res.append({
                    'pipeline': msg.pipeline_name,
                    'job': msg.job_id,
                    'ts': to_timestamp(msg.ts.seconds, msg.ts.nanos),
                    'message': message,
                    'worker': msg.worker_id,
                    'datum': msg.datum_id,
                    'user': msg.user,
                })
        return pd.DataFrame(res, columns=[
            'pipeline', 'job', 'ts', 'message',
            'worker', 'datum', 'user'
        ]).astype({
            'ts': 'datetime64[ns, UTC]',
            'user': 'bool',
        })

    @retry
    def inspect_pipeline(self, pipeline: str) -> Dict[str, Any]:
        res = self.pps_stub.InspectPipeline(InspectPipelineRequest(pipeline=Pipeline(name=pipeline)))
        info = dict(json.loads(MessageToJson(res)))
        for k in ['version', 'maxQueueSize', 'datumTries']:
            if k in info:
                info[k] = int(info[k])
        for k in ['createdAt']:
            if k in info:
                info[k] = pd.to_datetime(info[k]).to_pydatetime(warn=False)
        if 'input' in info:
            info['input'] = self._transform_cron_start(info['input'], lambda x: pd.to_datetime(x).to_pydatetime(warn=False))
        if 'state' in info:
            info['state'] = pipeline_state_mapping[info['state']]
        if 'jobCounts' in info:
            info['jobCounts'] = {job_state_mapping[int(k)]: v for k, v in info['jobCounts'].items()}
        if 'lastJobState' in info:
            info['lastJobState'] = job_state_mapping[info['lastJobState']]
        return info

    @retry
    def inspect_job(self, job: str) -> Dict[str, Any]:
        res = self.pps_stub.InspectJob(InspectJobRequest(job=Job(id=job)))
        info = dict(json.loads(MessageToJson(res)))
        for k in ['pipelineVersion', 'dataProcessed', 'dataSkipped', 'dataTotal', 'datumTries']:
            if k in info:
                info[k] = int(info[k])
        for k in ['started', 'finished']:
            if k in info:
                info[k] = pd.to_datetime(info[k]).to_pydatetime(warn=False)
        if 'input' in info:
            info['input'] = self._transform_cron_start(info['input'], lambda x: pd.to_datetime(x).to_pydatetime(warn=False))
        if 'state' in info:
            info['state'] = job_state_mapping[info['state']]
        if 'stats' in info:
            for k in ['downloadTime', 'processTime', 'uploadTime']:
                if k in info['stats']:
                    info['stats'][k] = float(info['stats'][k][:-1])
        return info

    @retry
    def inspect_datum(self, job: str, datum: str) -> Dict[str, Any]:
        res = self.pps_stub.InspectDatum(InspectDatumRequest(datum=Datum(id=datum, job=Job(id=job))))
        info = dict(json.loads(MessageToJson(res)))
        if 'state' in info:
            info['state'] = datum_state_mapping[info['state']]
        if 'stats' in info:
            for k in ['downloadTime', 'processTime', 'uploadTime']:
                if k in info['stats']:
                    info['stats'][k] = float(info['stats'][k][:-1])
        for file in info['data']:
            file['committed'] = pd.to_datetime(file['committed']).to_pydatetime(warn=False)
            file['fileType'] = file_type_mapping[file['fileType']]
        return info

    @retry
    def create_pipeline(self, pipeline_spec: dict) -> None:
        fields = {f.name for f in CreatePipelineRequest.DESCRIPTOR.fields}
        transform_fields = {f.name for f in Transform.DESCRIPTOR.fields}
        pipeline_spec = {k: v for k, v in pipeline_spec.items() if k in fields}
        pipeline_spec['transform'] = {k: v for k, v in pipeline_spec['transform'].items() if k in transform_fields}

        def protobuf_timestamp(x):
            dt = pd.to_datetime(x)
            if dt.tzinfo:
                dt = dt.tz_convert('utc').tz_localize(None)
            ts = Timestamp()
            ts.FromDatetime(dt.to_pydatetime())
            return ts

        pipeline_spec['input'] = self._transform_cron_start(pipeline_spec['input'], protobuf_timestamp)
        self.pps_stub.CreatePipeline(CreatePipelineRequest(**pipeline_spec))

    def update_pipeline(self, pipeline_spec: dict, reprocess: bool = False) -> None:
        pipeline_spec['update'] = True
        pipeline_spec['reprocess'] = reprocess
        self.create_pipeline(pipeline_spec)

    @retry
    def delete_pipeline(self, pipeline: str) -> None:
        self.pps_stub.DeletePipeline(DeletePipelineRequest(pipeline=Pipeline(name=pipeline)))

    @retry
    def start_pipeline(self, pipeline: str) -> None:
        self.pps_stub.StartPipeline(StartPipelineRequest(pipeline=Pipeline(name=pipeline)))

    @retry
    def stop_pipeline(self, pipeline: str) -> None:
        self.pps_stub.StopPipeline(StopPipelineRequest(pipeline=Pipeline(name=pipeline)))

    @retry
    def get_pipeline_cron_specs(self, pipeline: str) -> List[Dict[str, Any]]:
        res = self.pps_stub.InspectPipeline(InspectPipelineRequest(pipeline=Pipeline(name=pipeline)))
        return list(self._get_pipeline_input_cron_specs(res.input))

    @retry
    def create_repo(self, repo: str, description: Optional[str] = None) -> None:
        self.pfs_stub.CreateRepo(CreateRepoRequest(repo=Repo(name=repo), description=description))

    @retry
    def delete_repo(self, repo: str, force: bool = False) -> None:
        self.pfs_stub.DeleteRepo(DeleteRepoRequest(repo=Repo(name=repo), force=force))

    @retry
    def delete_commit(self, repo: str, commit: str) -> None:
        self.pfs_stub.DeleteCommit(DeleteCommitRequest(commit=Commit(repo=Repo(name=repo), id=commit)))

    @retry
    def create_branch(self, repo: str, commit: str, branch: str) -> None:
        repo = Repo(name=repo)
        commit = Commit(repo=repo, id=commit)
        branch = Branch(repo=repo, name=branch)
        self.pfs_stub.CreateBranch(CreateBranchRequest(head=commit, branch=branch))

    @retry
    def delete_branch(self, repo: str, branch: str) -> None:
        self.pfs_stub.DeleteBranch(DeleteBranchRequest(branch=Branch(repo=Repo(name=repo), name=branch)))

    @retry
    def delete_job(self, job: str) -> None:
        self.pps_stub.DeleteJob(DeleteJobRequest(job=Job(id=job)))

    @retry
    def get_file(self, repo: str, path: str, branch: Optional[str] = 'master', commit: Optional[str] = None) -> Generator[bytes, None, None]:
        if commit is None and branch is not None:
            commit = self.list_branch_heads(repo).get(branch)
        file = File(commit=Commit(repo=Repo(name=repo), id=commit), path=path)
        response = self.pfs_stub.GetFile(GetFileRequest(file=file))
        for content in response:
            yield content.value

    @retry
    def get_version(self) -> str:
        version = self.version_stub.GetVersion(Version())
        return f'{version.major}.{version.minor}.{version.micro}'

    @classmethod
    def _get_pipeline_input_cron_specs(cls, i: Input) -> Generator[Dict[str, Any], None, None]:
        if i.cron.spec != '':
            yield {
                'name': str(i.cron.name),
                'spec': str(i.cron.spec),
                'repo': str(i.cron.repo),
                'overwrite': bool(i.cron.overwrite),
            }
        cross_or_union = i.cross or i.union
        if cross_or_union:
            for j in cross_or_union:
                yield from cls._get_pipeline_input_cron_specs(j)

    @classmethod
    def _transform_cron_start(cls, i: dict, transformer: Callable) -> dict:
        for k, v in i.items():
            if k == 'cron' and 'start' in v:
                v['start'] = transformer(v['start'])
            elif k in ('cross', 'union'):
                for j in v:
                    j = cls._transform_cron_start(j, transformer)
        return i


class PachydermCommitAdapter:

    def __init__(self, adapter: PachydermAdapter, repo: str, branch: Optional[str] = 'master', parent_commit: Optional[str] = None, flush: bool = False):
        self.adapter = adapter
        self.repo = repo
        self.branch = branch
        self.parent_commit = parent_commit
        self.flush_ = flush
        self._commit = None
        self._finished = False
        self._buffer_size = 3 * 1024 * 1024  # 3 MB

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.delete()
        else:
            self.finish()
            if self.flush_:
                self.flush()

    @property
    def commit(self) -> Optional[str]:
        """Commit ID."""
        if self._commit is not None:
            return str(self._commit.id)
        else:
            return None

    @property
    def finished(self) -> bool:
        """Whether the commit is finished."""
        return self._finished

    @retry
    def start(self):
        """Start the commit."""
        self._raise_if_finished()
        parent_commit = Commit(repo=Repo(name=self.repo), id=self.parent_commit)
        self._commit = self.adapter.pfs_stub.StartCommit(StartCommitRequest(parent=parent_commit, branch=self.branch))

    @retry
    def finish(self):
        """Finish the commit."""
        self._raise_if_finished()
        self.adapter.pfs_stub.FinishCommit(FinishCommitRequest(commit=self._commit))
        self._finished = True

    @retry
    def flush(self, to_repos=None):
        """Blocks until all jobs triggered by this commit have finished.

        Args:
            to_repos: If specified, only the commits up to and including those repos
                will be considered, otherwise all repos are considered.
        """
        if to_repos:
            to_repos = [to_repos] if isinstance(to_repos, str) else to_repos
            to_repos = [Repo(name=repo) for repo in to_repos]
        res = self.adapter.pfs_stub.FlushCommit(FlushCommitRequest(commits=[self._commit], to_repos=to_repos or []))
        for _ in res:
            if res.done():
                return
            time.sleep(0.25)

    @retry
    def delete(self):
        """Delete the commit."""
        self.adapter.pfs_stub.DeleteCommit(DeleteCommitRequest(commit=self._commit))
        self._finished = True

    @retry
    def put_file_bytes(self, value: Union[str, bytes], path: str, encoding: str = 'utf-8',
                       delimiter: Optional[str] = None, target_file_datums: int = 0, target_file_bytes: int = 0) -> None:
        """Uploads a string or bytes `value` to a file in the given `path` in PFS.

        Args:
            value: The value to upload. If a string is given, it will be encoded using `encoding` (default UTF-8).
            path: PFS path to upload file to. Needs to be the full PFS path including filename.
            delimiter: Causes data to be broken up into separate files with `path` as a prefix.
                Possible values are 'none' (default), 'json', 'line', 'sql' and 'csv'.
            target_file_datum: Specifies the target number of datums in each written file.
                It may be lower if data does not split evenly, but will never be higher, unless the value is 0 (default).
            target_file_bytes: Specifies the target number of bytes in each written file.
                Files may have more or fewer bytes than the target.
        """
        self._raise_if_finished()
        if isinstance(value, str):
            value = value.encode(encoding)
        delimiter = {
            None: DELIMITER_NONE,
            'none': DELIMITER_NONE,
            'json': DELIMITER_JSON,
            'line': DELIMITER_LINE,
            'sql': DELIMITER_SQL,
            'csv': DELIMITER_CSV
        }[delimiter.lower() if delimiter is not None else None]
        file = File(commit=self._commit, path=path)

        def _blocks(v):
            for i in range(0, len(v) + 1, self._buffer_size):
                yield PutFileRequest(file=file, value=v[i:i + self._buffer_size], delimiter=delimiter,
                                     target_file_datums=target_file_datums, target_file_bytes=target_file_bytes)

        self.adapter.pfs_stub.PutFile(_blocks(value))

    @retry
    def put_file_url(self, url: str, path: str, recursive: bool = False) -> None:
        """Uploads a file using the content found at a URL.

        The URL is sent to the server which performs the request.

        Args:
            url: The URL to download content from.
            path: PFS path to upload file to. Needs to be the full PFS path including filename.
            recursive: Allow recursive scraping of some URL types, e.g. on s3:// URLs.
        """
        self._raise_if_finished()
        file = File(commit=self._commit, path=path)
        self.adapter.pfs_stub.PutFile(iter([PutFileRequest(file=file, url=url, recursive=recursive)]))

    @retry
    def delete_file(self, path: str) -> None:
        """Deletes the file found in a given `path` in PFS, if it exists.

        Args:
            path: PFS path of file to delete.
        """
        self._raise_if_finished()
        self.adapter.pfs_stub.DeleteFile(DeleteFileRequest(file=File(commit=self._commit, path=path)))

    @retry
    def create_branch(self, branch: str) -> None:
        """Sets this commit as a branch.

        Args:
            branch: Name of the branch.
        """
        branch = Branch(repo=Repo(name=self.repo), name=branch)
        self.adapter.pfs_stub.CreateBranch(CreateBranchRequest(head=self._commit, branch=branch))

    @retry
    def _list_file_paths(self, glob: str) -> List[str]:
        return [str(f.file.path) for f in self.adapter.pfs_stub.GlobFileStream(GlobFileRequest(commit=self._commit, pattern=glob))]

    def _raise_if_finished(self):
        if self.finished:
            raise PachydermError(f'Commit {self.commit} is already finished')



import os

import pandas as pd


def test_wildcard_filter():
    from pachypy.utils import wildcard_filter, wildcard_match
    x = ['a', 'ab', 'b']
    assert wildcard_match('', None) is True
    assert wildcard_match('', '*') is True
    assert wildcard_match('', '?') is False
    assert wildcard_match('a', '*') is True
    assert wildcard_match('a', 'a') is True
    assert wildcard_match('b', 'a') is False
    assert wildcard_filter(x, '*') == x
    assert wildcard_filter(x, None) == x
    assert wildcard_filter(x, 'a') == ['a']
    assert wildcard_filter(x, ['a']) == ['a']
    assert wildcard_filter(x, [['a']]) == ['a']
    assert wildcard_filter(x, 'a*') == ['a', 'ab']
    assert wildcard_filter(x, 'a?') == ['ab']
    assert wildcard_filter(x, [['a*'], 'b']) == x
    assert wildcard_filter(x, ['*a', '*b']) == x
    assert wildcard_filter(x, ['a', 'b']) == ['a', 'b']


def test_expand_files():
    from pathlib import Path
    from pachypy.utils import expand_files
    mock_dir = lambda glob: os.path.join(os.path.dirname(__file__), 'mock', glob)
    assert len(expand_files(None)) == 0
    assert len(expand_files(mock_dir('*.csv'))) == 7
    assert len(expand_files(Path(mock_dir('*.csv')))) == 7
    assert len(expand_files([mock_dir('list_*.csv'), Path(mock_dir('get_*.csv'))])) == 7


def test_invert_dict():
    from pachypy.utils import invert_dict
    assert invert_dict({'a': '1'}) == {'1': ['a']}
    assert invert_dict({'a': '1', 'b': '1'}) == {'1': ['a', 'b']}
    assert invert_dict({'a': '1', 'b': '2'}) == {'1': ['a'], '2': ['b']}


def test_to_timestamp():
    from pachypy.utils import to_timestamp
    assert to_timestamp(0, 0) is None
    assert to_timestamp(1554878996, 0) == pd.Timestamp('2019-04-10 06:49:56', tz='utc')


def test_to_timedelta():
    from pachypy.utils import to_timedelta
    assert to_timedelta(0, 0) == pd.Timedelta(0, unit='s')
    assert to_timedelta(0, 1) == pd.Timedelta(0.1, unit='s')



__all__ = [
    'PrettyPachydermClient'
]

import logging
import re
from typing import Dict, List, Iterable, Union, Optional
from datetime import datetime
from dateutil.relativedelta import relativedelta

import pandas.io.formats.style as style
import pandas as pd
import numpy as np
import yaml
from IPython.core.display import HTML
from termcolor import cprint
from tqdm import tqdm_notebook

from .client import PachydermClient, WildcardFilter


FONT_AWESOME_CSS_URL = 'https://use.fontawesome.com/releases/v5.8.1/css/all.css'
CLIPBOARD_JS_URL = 'https://cdnjs.cloudflare.com/ajax/libs/clipboard.js/2.0.4/clipboard.js'
BAR_COLOR = '#105ecd33'
PROGRESS_BAR_COLOR = '#03820333'

# Make yaml.dump() keep the order of keys in dictionaries
yaml.add_representer(
    dict,
    lambda self,
    data: yaml.representer.SafeRepresenter.represent_dict(self, data.items())  # type: ignore
)


def _fa(i: str) -> str:
    return f'<i class="fas fa-fw fa-{i}"></i>&nbsp;'


class CPrintHandler(logging.StreamHandler):

    def emit(self, record: logging.LogRecord):
        color = {
            logging.INFO: 'green',
            logging.WARNING: 'yellow',
            logging.ERROR: 'red',
            logging.CRITICAL: 'red',
        }.get(record.levelno, 'grey')
        cprint(self.format(record), color=color)


class PrettyTable(HTML):

    def __init__(self, styler: style.Styler, df: pd.DataFrame):
        super().__init__(data=styler.render())
        self.raw = df
        self.inject_dependencies()

    def inject_dependencies(self) -> None:
        fa_css = f'<link rel="stylesheet" href="{FONT_AWESOME_CSS_URL}" crossorigin="anonymous">'
        cb_js = f'''
            <script src="{CLIPBOARD_JS_URL}" crossorigin="anonymous"></script>
            <script>var clipboard = new ClipboardJS('.copyable');</script>
        '''
        self.data = fa_css + cb_js + self.data  # type: ignore


class PrettyYAML(HTML):

    def __init__(self, obj: object):
        super().__init__(data=self.format_yaml(obj))
        self.raw = obj

    @staticmethod
    def format_yaml(obj: object) -> str:
        s = str(yaml.dump(obj))
        s = re.sub(r'(^[\s-]*)([^\s]+:)', '\\1<span style="color: #888;">\\2</span>', s, flags=re.MULTILINE)
        return '<pre style="border: 1px #ccc solid; padding: 10px 12px; line-height: 140%;">' + s + '</pre>'


class PrettyPachydermClient(PachydermClient):

    table_styles = [
        dict(selector='th', props=[('text-align', 'left'), ('white-space', 'nowrap')]),
        dict(selector='td', props=[('text-align', 'left'), ('white-space', 'nowrap'), ('padding-right', '20px')]),
    ]

    @property
    def logger(self):
        if self._logger is None:
            self._logger = logging.getLogger('pachypy')
            self._logger.handlers = [CPrintHandler()]
            self._logger.setLevel(logging.DEBUG)
            self._logger.propagate = False
        return self._logger

    def list_repos(self, repos: WildcardFilter = '*') -> PrettyTable:
        df = super().list_repos(repos=repos)
        dfr = df.copy()
        df.rename({
            'repo': 'Repo',
            'is_tick': 'Tick',
            'branches': 'Branches',
            'size_bytes': 'Size',
            'created': 'Created',
        }, axis=1, inplace=True)
        df['Tick'] = df['Tick'].map({True: _fa('stopwatch'), False: ''})
        df['Branches'] = df['Branches'].apply(', '.join)
        styler = df[['Repo', 'Tick', 'Branches', 'Size', 'Created']].style \
            .bar(subset=['Size'], color=BAR_COLOR, vmin=0) \
            .format({'Created': self._format_datetime, 'Size': self._format_size}) \
            .set_properties(subset=['Branches'], **{'white-space': 'normal !important'}) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def list_commits(self, repos: WildcardFilter, n: int = 10) -> PrettyTable:
        df = super().list_commits(repos=repos, n=n)
        dfr = df.copy()
        df.rename({
            'repo': 'Repo',
            'commit': 'Commit',
            'branches': 'Branch',
            'size_bytes': 'Size',
            'started': 'Started',
            'finished': 'Finished',
            'parent_commit': 'Parent Commit',
        }, axis=1, inplace=True)
        styler = df[['Repo', 'Commit', 'Branch', 'Size', 'Started', 'Finished', 'Parent Commit']].style \
            .bar(subset=['Size'], color=BAR_COLOR, vmin=0) \
            .format({
                'Commit': self._format_hash,
                'Parent Commit': self._format_hash,
                'Branch': ', '.join,
                'Started': self._format_datetime,
                'Finished': self._format_datetime,
                'Size': self._format_size
            }) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def list_files(self, repos: WildcardFilter, branch: Optional[str] = 'master', commit: Optional[str] = None,
                   glob: str = '**', files_only: bool = True) -> PrettyTable:
        df = super().list_files(repos=repos, branch=branch, commit=commit, glob=glob, files_only=files_only)
        dfr = df.copy()
        df.rename({
            'repo': 'Repo',
            'type': 'Type',
            'path': 'Path',
            'size_bytes': 'Size',
            'commit': 'Commit',
            'branches': 'Branch',
            'committed': 'Committed',
        }, axis=1, inplace=True)
        styler = df[['Repo', 'Commit', 'Branch', 'Type', 'Path', 'Size', 'Committed']].style \
            .bar(subset=['Size'], color=BAR_COLOR, vmin=0) \
            .format({
                'Type': self._format_file_type,
                'Size': self._format_size,
                'Commit': self._format_hash,
                'Branch': ', '.join,
                'Committed': self._format_datetime
            }) \
            .set_properties(subset=['Path'], **{'white-space': 'normal !important'}) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def list_pipelines(self, pipelines: WildcardFilter = '*') -> PrettyTable:
        df = super().list_pipelines(pipelines=pipelines)
        dfr = df.copy()
        df['sort_key'] = df.index.map(self._calc_pipeline_sort_key(df['input_repos'].to_dict()))
        df.sort_values('sort_key', inplace=True)
        df.rename({
            'pipeline': 'Pipeline',
            'state': 'State',
            'cron_spec': 'Cron',
            'cron_prev_tick': 'Last Tick',
            'cron_next_tick': 'Next Tick',
            'input': 'Input',
            'output_branch': 'Output',
            'datum_tries': 'Tries',
            'created': 'Created',
        }, axis=1, inplace=True)
        df.loc[df['jobs_running'] > 0, 'State'] = 'job running'
        now = datetime.now(self.user_timezone)
        df['Next Tick In'] = (now - df['Next Tick']).dt.total_seconds() * -1
        df['Parallelism'] = ''
        df.loc[df['parallelism_constant'] > 0, 'Parallelism'] = \
            _fa('hashtag') + df['parallelism_constant'].astype(str)
        df.loc[df['parallelism_coefficient'] > 0, 'Parallelism'] = \
            _fa('asterisk') + df['parallelism_coefficient'].astype(str)
        df['Jobs'] = \
            '<span style="color: green">' + df['jobs_success'].astype(str) + '</span>' + \
            np.where(df['jobs_failure'] > 0, ' + <span style="color: red">' + df['jobs_failure'].astype(str) + '</span>', '')
        styler = df[['Pipeline', 'State', 'Cron', 'Next Tick In', 'Input', 'Output', 'Parallelism', 'Jobs', 'Created']].style \
            .apply(self._style_pipeline_state, subset=['State']) \
            .format({
                'State': self._format_pipeline_state,
                'Cron': self._format_cron_spec,
                'Next Tick In': self._format_duration,
                'Created': self._format_datetime,
            }) \
            .set_properties(subset=['Input'], **{'white-space': 'normal !important'}) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def list_jobs(self, pipelines: WildcardFilter = '*', n: int = 20, hide_null_jobs: bool = True) -> PrettyTable:
        df = super().list_jobs(pipelines=pipelines, n=n, hide_null_jobs=hide_null_jobs)
        dfr = df.copy()
        df.rename({
            'job': 'Job',
            'pipeline': 'Pipeline',
            'state': 'State',
            'started': 'Started',
            'duration': 'Duration',
            'restart': 'Restarts',
            'download_bytes': 'Downloaded',
            'upload_bytes': 'Uploaded',
            'output_commit': 'Output Commit',
        }, axis=1, inplace=True)
        df['Duration'] = df['Duration'].dt.total_seconds()
        df['Progress'] = \
            df['progress'].fillna(0).apply(lambda x: f'{x:.0%}') + ' | ' + \
            '<span style="color: green">' + df['data_processed'].astype(str) + '</span>' + \
            np.where(df['data_skipped'] > 0, ' + <span style="color: purple">' + df['data_skipped'].astype(str) + '</span>', '') + \
            ' / <span>' + df['data_total'].astype(str) + '</span>'
        styler = df[['Job', 'Pipeline', 'State', 'Started', 'Duration', 'Progress', 'Restarts', 'Downloaded', 'Uploaded', 'Output Commit']].style \
            .bar(subset=['Duration'], color=BAR_COLOR, vmin=0) \
            .apply(self._style_job_state, subset=['State']) \
            .apply(self._style_job_progress, subset=['Progress']) \
            .format({
                'Job': self._format_hash,
                'State': self._format_job_state,
                'Started': self._format_datetime,
                'Duration': self._format_duration,
                'Restarts': lambda i: _fa('undo') + str(i) if i > 0 else '',
                'Downloaded': self._format_size,
                'Uploaded': self._format_size,
                'Output Commit': self._format_hash
            }) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def list_datums(self, job: str) -> PrettyTable:
        df = super().list_datums(job=job)
        dfr = df.copy()
        df.rename({
            'job': 'Job',
            'datum': 'Datum',
            'state': 'State',
            'repo': 'Repo',
            'type': 'Type',
            'path': 'Path',
            'size_bytes': 'Size',
            'commit': 'Commit',
            'committed': 'Committed',
        }, axis=1, inplace=True)
        styler = df[['Job', 'Datum', 'State', 'Repo', 'Type', 'Path', 'Size', 'Commit', 'Committed']].style \
            .bar(subset=['Size'], color=BAR_COLOR, vmin=0) \
            .apply(self._style_datum_state, subset=['State']) \
            .format({
                'Job': self._format_hash,
                'Datum': self._format_hash,
                'State': self._format_datum_state,
                'Type': self._format_file_type,
                'Size': self._format_size,
                'Commit': self._format_hash,
                'Committed': self._format_datetime
            }) \
            .set_properties(subset=['Path'], **{'white-space': 'normal !important'}) \
            .set_table_styles(self.table_styles) \
            .hide_index()
        return PrettyTable(styler, dfr)

    def get_logs(self, pipelines: WildcardFilter = '*', datum: Optional[str] = None,
                 last_job_only: bool = True, user_only: bool = False, master: bool = False) -> None:
        df = super().get_logs(pipelines=pipelines, last_job_only=last_job_only, user_only=user_only, master=master)
        job = None
        worker = None
        for _, row in df.iterrows():
            if row.job != job:
                print()
                cprint(f' Pipeline {row.pipeline} | Job {row.job} ', 'yellow', 'on_grey')
            if row.worker != worker:
                cprint(f' Worker {row.worker} ', 'white', 'on_grey')
            color = 'grey' if row.user else 'blue'
            message = row.message
            if 'warning' in message.lower():
                color = 'magenta'
            elif 'error' in message.lower() or 'exception' in message.lower() or 'critical' in message.lower():
                color = 'red'
            cprint(f'[{row.ts}] {message}', color)
            job = row.job
            worker = row.worker

    def inspect_pipeline(self, pipeline: str) -> PrettyYAML:
        info = super().inspect_pipeline(pipeline)
        return PrettyYAML(info)

    def inspect_job(self, job: str) -> PrettyYAML:
        info = super().inspect_job(job)
        return PrettyYAML(info)

    def inspect_datum(self, job: str, datum: str) -> PrettyYAML:
        info = super().inspect_datum(job, datum)
        return PrettyYAML(info)

    @staticmethod
    def _calc_pipeline_sort_key(input_repos: Dict[str, List[str]]):
        def get_dag_distance(p, i=0):
            yield i
            for d in input_repos[p]:
                if d in pipelines:
                    yield from get_dag_distance(d, i + 1)

        def get_dag_dependencies(p):
            yield p
            for d in input_repos[p]:
                if d in pipelines:
                    yield from get_dag_dependencies(d)

        pipelines = set(input_repos.keys())
        dag_distance = {p: max(list(get_dag_distance(p))) for p in pipelines}
        dag_nodes = {p: set(get_dag_dependencies(p)) for p in pipelines}
        for p, nodes in dag_nodes.items():
            for node in nodes:
                dag_nodes[node].update(nodes)
        dag_name = {p: min(nodes) for p, nodes in dag_nodes.items()}
        return {p: f'{dag_name[p]}/{dag_distance[p]}' for p in pipelines}

    def _format_datetime(self, d: datetime) -> str:
        if pd.isna(d):
            return ''
        td = (datetime.now(self.user_timezone).date() - d.date()).days
        word = {-1: 'Tomorrow', 0: 'Today', 1: 'Yesterday'}
        return (word[td] if td in word else f'{d:%-d %b %Y}') + f' at {d:%H:%M}'

    @staticmethod
    def _format_duration(secs: float, n: int = 2) -> str:
        if pd.isna(secs):
            return ''
        d = relativedelta(seconds=int(secs), microseconds=int((secs % 1) * 1e6))
        attrs = {
            'years': 'years',
            'months': 'months',
            'days': 'days',
            'hours': 'hours',
            'minutes': 'mins',
            'seconds': 'secs',
            'microseconds': 'ms'
        }
        ret = ''
        i = 0
        for attr, attr_short in attrs.items():
            x = getattr(d, attr, 0)
            if x > 0:
                if attr == 'microseconds':
                    x /= 1000
                    u = attr_short
                else:
                    u = x != 1 and attr_short or attr_short[:-1]
                ret += f'{x:.0f} {u}, '
                i += 1
                if i >= n or attr in {'minutes', 'seconds'}:
                    break
        return ret.strip(', ')

    @staticmethod
    def _format_size(x: Union[int, float]) -> str:
        if abs(x) == 1:
            return f'{x:.0f} byte'
        if abs(x) < 1000.0:
            return f'{x:.0f} bytes'
        x /= 1000.0
        for unit in ['KB', 'MB', 'GB', 'TB']:
            if abs(x) < 1000.0:
                return f'{x:.1f} {unit}'
            x /= 1000.0
        return f'{x:,.1f} PB'

    @staticmethod
    def _format_hash(s: str) -> str:
        if pd.isna(s):
            return ''
        short = s[:5] + '..' + s[-5:] if len(s) > 12 else s
        return f'<pre class="copyable" title="{s} (click to copy)" data-clipboard-text="{s}" style="cursor: copy; background: none; white-space: nowrap;">{short}</pre>'

    @staticmethod
    def _format_cron_spec(s: str) -> str:
        if pd.isna(s) or s == '':
            return ''
        return _fa('stopwatch') + s

    @staticmethod
    def _format_file_type(s: str) -> str:
        return {
            'file': _fa('file') + s,
            'dir': _fa('folder') + s,
        }.get(s, s)

    @staticmethod
    def _format_pipeline_state(s: str) -> str:
        return {
            'starting': _fa('spinner') + s,
            'restarting': _fa('undo') + s,
            'running': _fa('toggle-on') + s,
            'job running': _fa('running') + s,
            'failure': _fa('bolt') + s,
            'paused': _fa('toggle-off') + s,
            'standby': _fa('power-off') + s,
        }.get(s, s)

    @staticmethod
    def _format_job_state(s: str) -> str:
        return {
            'unknown': _fa('question') + s,
            'starting': _fa('spinner') + s,
            'running': _fa('running') + s,
            'merging': _fa('compress-arrows-alt') + s,
            'success': _fa('check') + s,
            'failure': _fa('bolt') + s,
            'killed': _fa('skull-crossbones') + s,
        }.get(s, s)

    @staticmethod
    def _format_datum_state(s: str) -> str:
        return {
            'unknown': _fa('question') + s,
            'starting': _fa('spinner') + s,
            'skipped': _fa('forward') + s,
            'success': _fa('check') + s,
            'failed': _fa('bolt') + s,
        }.get(s, s)

    @staticmethod
    def _style_pipeline_state(s: Iterable[str]) -> List[str]:
        color = {
            'starting': 'orange',
            'restarting': 'orange',
            'running': 'green',
            'job running': 'purple',
            'failure': 'red',
            'paused': 'orange',
            'standby': '#0251c9',
        }
        return [f"color: {color.get(v, 'gray')}; font-weight: bold" for v in s]

    @staticmethod
    def _style_job_state(s: Iterable[str]) -> List[str]:
        color = {
            'starting': 'orange',
            'running': 'orange',
            'merging': 'orange',
            'success': 'green',
            'failure': 'red',
            'killed': 'red',
        }
        return [f"color: {color.get(v, 'gray')}; font-weight: bold" for v in s]

    @staticmethod
    def _style_datum_state(s: Iterable[str]) -> List[str]:
        color = {
            'starting': 'orange',
            'skipped': '#0251c9',
            'success': 'green',
            'failed': 'red',
        }
        return [f"color: {color.get(v, 'gray')}; font-weight: bold" for v in s]

    @staticmethod
    def _style_job_progress(s: pd.Series) -> List[str]:
        def css_bar(end):
            css = 'width: 10em; height: 80%;'
            if end > 0:
                css += 'background: linear-gradient(90deg,'
                css += '{c} {e:.1f}%, transparent {e:.1f}%)'.format(e=min(end, 100), c=PROGRESS_BAR_COLOR)
            return css
        s = s.apply(lambda x: float(x.split('%')[0]))
        return [css_bar(x) if not pd.isna(x) and x < 100 else '' for x in s]

    @classmethod
    def _progress(cls, x, n=None, **kwargs):
        n = n or len(x)
        if 'leave' not in kwargs:
            kwargs['leave'] = False
        return tqdm_notebook(x, **kwargs) if n > 2 else x



__all__ = [
    'DockerRegistryAdapter', 'AmazonECRAdapter'
]

import re
from base64 import b64decode
from requests.exceptions import HTTPError
from typing import Tuple, Optional

import boto3
import docker


class RegistryException(Exception):
    pass


class DockerRegistryAdapter:

    """Docker registry adapter."""

    def __init__(self, docker_client: docker.DockerClient):
        self.docker_client = docker_client

    def get_image_digest(self, image: str) -> str:
        # TODO: This can be simplified by using docker.api.APIClient.inspect_distribution()
        # once a fixed version is released.

        api = self.docker_client.api
        registry, _ = docker.auth.resolve_repository_name(image)
        header = docker.auth.get_config_header(api, registry)

        try:
            url = api._url("/distribution/{0}/json", image)
            headers = {'X-Registry-Auth': header} if header else {}
            result = api._result(api._get(url, headers=headers), True)
            return str(result['Descriptor']['digest'])
        except HTTPError:
            raise RegistryException(f"Image '{image}' not found in Docker registry '{registry}'")

    def push_image(self, image) -> str:
        digest = None
        push_stream = self.docker_client.api.push(repository=image, stream=True, decode=True)
        for chunk in push_stream:
            if 'error' in chunk:
                raise RegistryException(chunk['error'])
            if 'aux' in chunk:
                digest = str(chunk['aux']['Digest'])
        if not digest:
            raise RegistryException('did not retrieve image digest from Docker after pushing image')
        return digest


class AmazonECRAdapter(DockerRegistryAdapter):

    """Amazon Elastic Container Registry (ECR) adapter using boto3.

    Getting image digests via boto3 is faster than using the DockerRegistryAdapter,
    which otherwise works fine for ECR too.
    When pushing images, this class also handles logging in to ECR by getting an
    authorization token using boto3 and passing it to Docker to log in.

    Args:
        aws_access_key_id: AWS access key ID.
        aws_secret_access_key: AWS secret access key.
    """

    def __init__(self, docker_client: docker.DockerClient):
        self.docker_client = docker_client
        self.ecr_client = boto3.client('ecr')

    def set_credentials(self, aws_access_key_id: str, aws_secret_access_key: str):
        self.ecr_client = boto3.client('ecr', aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)

    def login(self, registry: str) -> None:
        response = self.ecr_client.get_authorization_token(registryIds=[registry[:12]])
        auth_token = response['authorizationData'][0]['authorizationToken']
        username, password = b64decode(auth_token).decode('utf-8').split(':')
        self.docker_client.login(username=username, password=password, registry=registry, reauth=True)

    def get_image_digest(self, image: str) -> str:
        registry, repository, tag = self._split_image_string(image)
        if not tag:
            raise ValueError('tag is required for images in Amazon ECR')
        try:
            res = self.ecr_client.batch_get_image(registryId=registry[:12], repositoryName=repository, imageIds=[{'imageTag': tag}])
            return str(res['images'][0]['imageId']['imageDigest'])
        except (self.ecr_client.exceptions.RepositoryNotFoundException, KeyError):
            raise RegistryException(f"Image '{repository}:{tag}' not found in Amazon ECR registry '{registry[:12]}'")

    def push_image(self, image: str) -> str:
        try:
            return super().push_image(image)
        except RegistryException as e:
            if 'denied' in str(e):
                registry = self._split_image_string(image)[0]
                self.login(registry)
                return super().push_image(image)
            else:
                raise e

    @staticmethod
    def _split_image_string(image) -> Tuple[str, Optional[str], Optional[str]]:
        match = re.match('(^[^:@]+)(?::([^:@]+))?(?:@(.+))?$', image)
        if match:
            repository, tag, _ = match.groups()
            registry, repository = docker.auth.resolve_repository_name(repository)
            return registry, repository, tag
        else:
            raise ValueError(f"'{image}' is not a valid image string")



import pytest

from pachypy.pretty import PrettyPachydermClient
from test_client import patch_adapter


@pytest.fixture(scope='module')
def pretty_client():
    return PrettyPachydermClient()


def test_list_repos(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_repos()
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 7
    with patch_adapter(empty=True):
        html = pretty_client.list_repos()
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_list_commits(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_commits('test_x_pipeline_3')
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 11
    with patch_adapter(empty=True):
        html = pretty_client.list_commits('test_x_pipeline_3')
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_list_files(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_files('test_x_pipeline_3', files_only=False)
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 8
    with patch_adapter(empty=True):
        html = pretty_client.list_files('test_x_pipeline_3', files_only=False)
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_list_pipelines(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_pipelines()
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 6
    with patch_adapter(empty=True):
        html = pretty_client.list_pipelines()
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_list_jobs(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_jobs()
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 9
    with patch_adapter(empty=True):
        html = pretty_client.list_jobs()
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_list_datums(pretty_client: PrettyPachydermClient):
    with patch_adapter():
        html = pretty_client.list_datums('e26ccf65131b4b3d9087cebc2f944279')
        assert 'use.fontawesome.com' in html.data
        assert '<table' in html.data and html.data.count('<tr') == 11
    with patch_adapter(empty=True):
        html = pretty_client.list_datums('e26ccf65131b4b3d9087cebc2f944279')
        assert '<table' in html.data and html.data.count('<tr') == 1


def test_get_logs(pretty_client: PrettyPachydermClient, capsys):
    with patch_adapter():
        pretty_client.get_logs('test_x_pipeline_5', last_job_only=False)
        output = capsys.readouterr().out
        assert output.count('\n') == 25
        assert output.count(' | Job ') == 2
    with patch_adapter(empty=True):
        pretty_client.get_logs('test_x_pipeline_5', last_job_only=False)
        assert output.count('\n') == 25


@patch_adapter()
def test_inspect_pipeline(pretty_client: PrettyPachydermClient, **mocks):
    del mocks
    html = pretty_client.inspect_pipeline('pipeline')
    assert '<pre' in html.data
    assert isinstance(html.raw, dict)


@patch_adapter()
def test_inspect_job(pretty_client: PrettyPachydermClient, **mocks):
    del mocks
    html = pretty_client.inspect_job('abcd1234')
    assert '<pre' in html.data
    assert isinstance(html.raw, dict)


@patch_adapter()
def test_inspect_datum(pretty_client: PrettyPachydermClient, **mocks):
    del mocks
    html = pretty_client.inspect_datum('abcd1234', 'abcdefgh12345678')
    assert '<pre' in html.data
    assert isinstance(html.raw, dict)


def test_pipeline_sort_key():
    from pachypy.pretty import PrettyPachydermClient
    assert PrettyPachydermClient._calc_pipeline_sort_key({'a': [], 'b': ['a']}) == {'a': 'a/0', 'b': 'a/1'}
    assert PrettyPachydermClient._calc_pipeline_sort_key({'b': [], 'a': ['b']}) == {'b': 'a/0', 'a': 'a/1'}
    assert PrettyPachydermClient._calc_pipeline_sort_key({'a': [], 'b': [], 'c': ['a', 'b'], 'd': []}) == {'a': 'a/0', 'b': 'a/0', 'c': 'a/1', 'd': 'd/0'}
    assert PrettyPachydermClient._calc_pipeline_sort_key({'a': [], 'b': ['a'], 'c': ['a'], 'd': ['b']}) == {'a': 'a/0', 'b': 'a/1', 'c': 'a/1', 'd': 'a/2'}


def test_style_job_progress():
    import pandas as pd
    assert not PrettyPachydermClient._style_job_progress(pd.Series(['0% | 0 + 0 / 2']))[0].endswith('0.0%)')
    assert PrettyPachydermClient._style_job_progress(pd.Series(['50% | 1 + 0 / 2']))[0].endswith('50.0%)')
    assert PrettyPachydermClient._style_job_progress(pd.Series(['100% | 1 + 1 / 2']))[0] == ''


def test_format_datetime(pretty_client):
    from datetime import datetime
    from dateutil.relativedelta import relativedelta as rd
    t = datetime.now(pretty_client.user_timezone)
    t = [t.year, t.month, t.day]
    assert pretty_client._format_datetime(None) == ''
    assert pretty_client._format_datetime(datetime(2019, 1, 2, 11, 11, 23)) == '2 Jan 2019 at 11:11'
    assert pretty_client._format_datetime(datetime(2019, 2, 3, 23, 11, 59)) == '3 Feb 2019 at 23:11'
    assert pretty_client._format_datetime(datetime(*t + [11, 11])) == 'Today at 11:11'
    assert pretty_client._format_datetime(datetime(*t + [11, 11]) - rd(days=1)) == 'Yesterday at 11:11'
    assert pretty_client._format_datetime(datetime(*t + [11, 11]) + rd(days=1)) == 'Tomorrow at 11:11'


def test_format_duration():
    assert PrettyPachydermClient._format_duration(None) == ''
    assert PrettyPachydermClient._format_duration(0) == ''
    assert PrettyPachydermClient._format_duration(0.001) == '1 ms'
    assert PrettyPachydermClient._format_duration(1) == '1 sec'
    assert PrettyPachydermClient._format_duration(2) == '2 secs'
    assert PrettyPachydermClient._format_duration(2.1) == '2 secs'
    assert PrettyPachydermClient._format_duration(60) == '1 min'
    assert PrettyPachydermClient._format_duration(120) == '2 mins'
    assert PrettyPachydermClient._format_duration(121) == '2 mins'
    assert PrettyPachydermClient._format_duration(3600) == '1 hour'
    assert PrettyPachydermClient._format_duration(7200) == '2 hours'
    assert PrettyPachydermClient._format_duration(3660) == '1 hour, 1 min'
    assert PrettyPachydermClient._format_duration(7260) == '2 hours, 1 min'
    assert PrettyPachydermClient._format_duration(7320) == '2 hours, 2 mins'
    assert PrettyPachydermClient._format_duration(7322) == '2 hours, 2 mins'
    assert PrettyPachydermClient._format_duration(86400) == '1 day'
    assert PrettyPachydermClient._format_duration(90000) == '1 day, 1 hour'


def test_format_size():
    assert PrettyPachydermClient._format_size(0) == '0 bytes'
    assert PrettyPachydermClient._format_size(1) == '1 byte'
    assert PrettyPachydermClient._format_size(1000) == '1.0 KB'
    assert PrettyPachydermClient._format_size(10100) == '10.1 KB'
    assert PrettyPachydermClient._format_size(1200000) == '1.2 MB'
    assert PrettyPachydermClient._format_size(100300000) == '100.3 MB'
    assert PrettyPachydermClient._format_size(1400000000) == '1.4 GB'
    assert PrettyPachydermClient._format_size(1500000000000) == '1.5 TB'
    assert PrettyPachydermClient._format_size(1600000000000000) == '1.6 PB'



__all__ = [
    'PachydermClient', 'PachydermCommit'
]

import io
import json
import logging
import os
import re
from collections import namedtuple
from datetime import datetime, tzinfo
from pathlib import Path
from typing import List, Set, Dict, Iterable, Callable, IO, BinaryIO, Union, Optional, Any

import chardet
import pandas as pd
import pytz
import yaml
from croniter import croniter
from docker import DockerClient
from tzlocal import get_localzone

from .adapter import PachydermAdapter, PachydermCommitAdapter, PachydermError
from .registry import DockerRegistryAdapter, AmazonECRAdapter
from .utils import WildcardFilter, FileGlob, wildcard_filter, wildcard_match, expand_files


PipelineChanges = namedtuple('PipelineChanges', ['created', 'updated', 'deleted'])


class PachydermClientError(PachydermError):

    def __init__(self, message):
        super().__init__(message)


class DockerError(Exception):
    pass


class PachydermClient:

    """Pachyderm client.

    Args:
        host: Hostname or IP address to reach pachd. Attempts to get this from the
            environment variable PACHD_ADDRESS or ``~/.pachyderm/config.json`` if not set.
        port: Port on which pachd is listening. Defaults to 30650.
        add_image_digests: Whether to add a digest to the image field in pipeline specs to
            to force Pachyderm to pull the latest version from the container registry.
        build_images: Whether to build Docker images for pipelines and push them
            to the container registry. Only applies to pipelines that have the `dockerfile`
            or `dockerfile_path` directive set within the `transform` field.
        pipeline_spec_files: Glob pattern(s) to pipeline spec files in YAML or JSON format.
        pipeline_spec_transformer: Function that takes a pipeline spec as dictionary
            as the only argument and returns a transformed pipeline spec.
        pachd_timezone: Timezone of the system that pachd is running on. This determines
            when cron inputs trigger. Defaults to UTC.
        user_timezone: Timezone of the user which is used to localize timestamps.
            Attempts to get this from the environment variable TZ or system settings if not specified.
    """

    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        add_image_digests: bool = True,
        build_images: bool = True,
        pipeline_spec_files: FileGlob = None,
        pipeline_spec_transformer: Optional[Callable[[dict], dict]] = None,
        pachd_timezone: Union[str, tzinfo] = 'utc',
        user_timezone: Optional[Union[str, tzinfo]] = None
    ):
        self._pipeline_spec_files: FileGlob = None
        self._docker_client: Optional[DockerClient] = None
        self._docker_registry: Optional[DockerRegistryAdapter] = None
        self._amazon_ecr: Optional[AmazonECRAdapter] = None
        self._image_digests: Dict[str, str] = {}
        self._built_images: Set[str] = set()
        self._logger: Optional[logging.Logger] = None
        self._user_timezone: Optional[tzinfo] = None
        self._pachd_timezone: Optional[tzinfo] = None

        self.adapter = PachydermAdapter(host=host, port=port)
        self.add_image_digests = add_image_digests
        self.build_images = build_images
        self.pipeline_spec_files = pipeline_spec_files
        self.pipeline_spec_transformer = pipeline_spec_transformer
        self.pachd_timezone = pachd_timezone
        self.user_timezone = user_timezone

        self.logger.debug(f'Created client for Pachyderm cluster at {self.adapter.host}:{self.adapter.port}')

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = logging.getLogger('pachypy')
        return self._logger

    @property
    def pipeline_spec_files(self) -> List[str]:
        return expand_files(self._pipeline_spec_files)

    @pipeline_spec_files.setter
    def pipeline_spec_files(self, files: FileGlob):
        self._pipeline_spec_files = files

    @property
    def docker_client(self) -> DockerClient:
        if self._docker_client is None:
            self._docker_client = DockerClient.from_env()
        return self._docker_client

    @docker_client.setter
    def docker_client(self, docker_client: DockerClient):
        self._docker_client = docker_client
        if self._docker_registry is not None:
            self._docker_registry.docker_client = docker_client
        if self._amazon_ecr is not None:
            self._amazon_ecr.docker_client = docker_client

    @property
    def docker_registry(self) -> DockerRegistryAdapter:
        if self._docker_registry is None:
            self._docker_registry = DockerRegistryAdapter(self.docker_client)
        return self._docker_registry

    @property
    def amazon_ecr(self) -> AmazonECRAdapter:
        if self._amazon_ecr is None:
            self._amazon_ecr = AmazonECRAdapter(self.docker_client)
        return self._amazon_ecr

    @property
    def user_timezone(self) -> tzinfo:
        if self._user_timezone is None:
            tz = os.environ.get('TZ')
            if tz is not None:
                self._user_timezone = pytz.timezone(tz)
            else:
                self._user_timezone = get_localzone()  # type: ignore
        return self._user_timezone

    @user_timezone.setter
    def user_timezone(self, tz: Optional[Union[tzinfo, str]]):
        if isinstance(tz, str):
            self._user_timezone = pytz.timezone(tz)
        else:
            self._user_timezone = tz

    @property
    def pachd_timezone(self):
        return self._pachd_timezone or pytz.utc

    @pachd_timezone.setter
    def pachd_timezone(self, tz: Optional[Union[tzinfo, str]]):
        if isinstance(tz, str):
            self._pachd_timezone = pytz.timezone(tz)
        else:
            self._pachd_timezone = tz

    @property
    def pachd_version(self) -> str:
        return self.adapter.get_version()

    def list_repos(self, repos: WildcardFilter = '*') -> pd.DataFrame:
        """Get list of repos as pandas DataFrame.

        Args:
            repos: Name pattern to filter repos returned. Supports shell-style wildcards.
        """
        df = self.adapter.list_repos()
        if repos is not None and repos != '*':
            df = df[df.repo.isin(set(wildcard_filter(df.repo, repos)))]
        df['is_tick'] = df['repo'].str.endswith('_tick')
        df['created'] = self._localize_timestamp(df['created'])
        df = df.sort_values(['repo']).reset_index(drop=True)
        return df[['repo', 'is_tick', 'branches', 'size_bytes', 'created']].astype({
            'is_tick': 'bool',
        })

    def list_commits(self, repos: WildcardFilter, n: int = 10) -> pd.DataFrame:
        """Get list of commits as pandas DataFrame.

        Args:
            repos: Name pattern to filter repos to return commits for. Supports shell-style wildcards.
            n: Maximum number of commits returned per repo.
        """
        repo_names = self._list_repo_names(repos)
        if len(repo_names) == 0:
            raise PachydermClientError(f'No repos matching "{repos}" were found.')
        df = pd.concat([self.adapter.list_commits(repo=repo, n=n) for repo in self._progress(repo_names, unit='repo')])
        for col in ('started', 'finished'):
            df[col] = self._localize_timestamp(df[col])
        df = df.sort_values(['repo', 'started'], ascending=[True, False]).reset_index(drop=True)
        return df[['repo', 'commit', 'branches', 'size_bytes', 'started', 'finished', 'parent_commit']]

    def list_files(self, repos: WildcardFilter, branch: Optional[str] = 'master', commit: Optional[str] = None,
                   glob: str = '**', files_only: bool = True) -> pd.DataFrame:
        """Get list of files as pandas DataFrame.

        Args:
            repos: Name pattern to filter repos to return files for. Supports shell-style wildcards.
            branch: Branch to list files for. Defaults to 'master'.
            commit: Commit ID to return files for. Overrides `branch` if specified.
                If specified, the repos parameter must only match the repo this commit ID belongs to.
            glob: Glob pattern to filter files returned.
            files_only: Whether to return only files or include directories.
        """
        repo_names = self._list_repo_names(repos)
        if len(repo_names) == 0:
            raise PachydermClientError(f'No repos matching "{repos}" were found.')
        if commit is not None and len(repo_names) > 1:
            raise PachydermClientError(f'More than one repo matches "{repos}", but it must only match the repo of commit ID "{commit}".')
        if branch is None and commit is None:
            df = pd.concat([
                self.adapter.list_files(repo=repo, branch=None, commit=branch_head, glob=glob)
                for repo in self._progress(repo_names, unit='repo') for branch_head in self.adapter.list_branch_heads(repo).values()
            ])
        else:
            df = pd.concat([
                self.adapter.list_files(repo=repo, branch=branch, commit=commit, glob=glob)
                for repo in self._progress(repo_names, unit='repo')
            ])
        if files_only:
            df = df[df['type'] == 'file']
        df['committed'] = self._localize_timestamp(df['committed'])
        df = df.sort_values(['repo', 'path']).reset_index(drop=True)
        return df[['repo', 'path', 'type', 'size_bytes', 'commit', 'branches', 'committed']]

    def list_pipelines(self, pipelines: WildcardFilter = '*') -> pd.DataFrame:
        """Get list of pipelines as pandas DataFrame.

        Args:
            pipelines: Name pattern to filter pipelines returned. Supports shell-style wildcards.
        """
        df = self.adapter.list_pipelines()
        if pipelines is not None and pipelines != '*':
            df = df[df.pipeline.isin(set(wildcard_filter(df.pipeline, pipelines)))]
        df['created'] = self._localize_timestamp(df['created'])
        df['cron_prev_tick'] = df['cron_spec'].apply(lambda cs: self._calc_cron_tick(cs, prev=True))
        df['cron_next_tick'] = df['cron_spec'].apply(lambda cs: self._calc_cron_tick(cs, prev=False))
        df = df.sort_values(['pipeline']).reset_index(drop=True)
        return df[[
            'pipeline', 'state', 'cron_spec', 'cron_prev_tick', 'cron_next_tick', 'input', 'input_repos', 'output_branch',
            'parallelism_constant', 'parallelism_coefficient', 'datum_tries', 'max_queue_size',
            'jobs_running', 'jobs_success', 'jobs_failure', 'created'
        ]].astype({
            'cron_prev_tick': f'datetime64[ns, {self.user_timezone}]',
            'cron_next_tick': f'datetime64[ns, {self.user_timezone}]',
        })

    def list_jobs(self, pipelines: WildcardFilter = '*', n: int = 20, hide_null_jobs: bool = True) -> pd.DataFrame:
        """Get list of jobs as pandas DataFrame.

        Args:
            pipelines: Pattern to filter jobs by pipeline name. Supports shell-style wildcards.
            n: Maximum number of jobs returned.
            hide_null_jobs: If true, empty jobs with no data to process will be filtered out.
        """
        if pipelines is not None and pipelines != '*':
            pipeline_names = self._list_pipeline_names(pipelines)
            if len(pipeline_names) == 0:
                raise PachydermClientError(f'No pipelines matching "{pipelines}" were found.')
            df = pd.concat([self.adapter.list_jobs(pipeline=pipeline, n=n) for pipeline in self._progress(pipeline_names, unit='pipeline')])
        else:
            df = self.adapter.list_jobs(n=n)
        if hide_null_jobs:
            df = df[df['data_total'] > 0]
        for col in ('started', 'finished'):
            df[col] = self._localize_timestamp(df[col])
        df = df.reset_index().sort_values(['started', 'index'], ascending=[False, True]).head(n).reset_index(drop=True)
        df['duration'] = df['finished'] - df['started']
        now = pd.Timestamp('now', tz=self.user_timezone)
        df.loc[df['state'] == 'running', 'duration'] = now - df['started']
        df['progress'] = (df['data_processed'] + df['data_skipped']) / df['data_total']
        return df[[
            'job', 'pipeline', 'state', 'started', 'finished', 'duration',
            'data_processed', 'data_skipped', 'data_total', 'progress', 'restart',
            'download_time', 'process_time', 'upload_time',
            'download_bytes', 'upload_bytes', 'output_commit',
        ]]

    def list_datums(self, job: str) -> pd.DataFrame:
        """Get list of datums for a job as pandas DataFrame.

        Args:
            job: Job ID to return datums for.
        """
        df = self.adapter.list_datums(job)
        df['committed'] = self._localize_timestamp(df['committed'])
        df = df.sort_values(['datum']).reset_index(drop=True)
        return df[[
            'datum', 'job', 'state', 'repo', 'path', 'type', 'size_bytes',
            'commit', 'committed'
        ]]

    def get_logs(self, pipelines: WildcardFilter = '*', datum: Optional[str] = None,
                 last_job_only: bool = True, user_only: bool = False, master: bool = False) -> pd.DataFrame:
        """Get logs for jobs.

        Args:
            pipelines: Pattern to filter logs by pipeline name. Supports shell-style wildcards.
            datum: If specified logs are filtered to a datum ID.
            last_job_only: Whether to only show/return logs for the last job of each pipeline.
            user_only: Whether to only return logs generated by user code.
            master: Whether to include logs from the master process.
        """
        pipeline_names = self._list_pipeline_names(pipelines)
        if len(pipeline_names) == 0:
            raise PachydermClientError(f'No pipelines matching "{pipelines}" were found.')
        logs = [self.adapter.get_logs(pipeline=pipeline, master=master) for pipeline in self._progress(pipeline_names, unit='pipeline')]
        df = pd.concat(logs, ignore_index=True).reset_index()
        df = df[df['job'].notna()]
        df['user'] = df['user'].fillna(False)
        if datum is not None:
            df = df[df['datum'] == datum]
        if user_only:
            df = df[df['user']]
        df['message'] = df['message'].fillna('')
        df['ts'] = self._localize_timestamp(df['ts'])
        df['worker_ts_min'] = df.groupby(['job', 'worker'])['ts'].transform('min')
        if last_job_only and len(df) > 0:
            df['job_ts_min'] = df.groupby(['job'])['ts'].transform('min')
            df['job_rank'] = df.groupby(['pipeline'])['job_ts_min'].transform(lambda x: x.rank(method='dense', ascending=False))
            df = df[df['job_rank'] == 1]
        df = df.sort_values(['worker_ts_min', 'job', 'worker', 'ts', 'index'], ascending=True).reset_index(drop=True)
        return df[['ts', 'job', 'pipeline', 'worker', 'datum', 'message', 'user']]

    def inspect_pipeline(self, pipeline: str) -> Dict[str, Any]:
        """Returns info about a pipeline.

        Args:
            pipeline: Name of pipeline to get info for.
        """
        return self.adapter.inspect_pipeline(pipeline=pipeline)

    def inspect_job(self, job: str) -> Dict[str, Any]:
        """Returns info about a job.

        Args:
            job: ID of job to get info for.
        """
        return self.adapter.inspect_job(job=job)

    def inspect_datum(self, job: str, datum: str) -> Dict[str, Any]:
        """Returns info about a datum.

        Only works if stats tracking is enabled for the pipeline (enable_stats)
        and raises a PachydermError otherwise.

        Args:
            job: ID of job the datum belongs to.
            datum: ID of datum to get info for.
        """
        return self.adapter.inspect_datum(job=job, datum=datum)

    def create_repos(self, repos: Union[str, Iterable[str]]) -> List[str]:
        """Create one or multiple new repositories in pfs.

        Args:
            repos: Name of new repository or iterable of names.

        Returns:
            Created repos.
        """
        if isinstance(repos, str):
            repos = [repos]
        existing_repos = set(self._list_repo_names())
        created_repos = []
        for repo in repos:
            if repo not in existing_repos:
                self.adapter.create_repo(repo=repo)
                created_repos.append(repo)
                self.logger.info(f'Created repo {repo}')
            else:
                self.logger.warning(f'Repo {repo} already exists')
        return created_repos

    def delete_repos(self, repos: WildcardFilter) -> List[str]:
        """Delete repositories.

        Args:
            repos: Pattern to filter repos to delete. Supports shell-style wildcards.

        Returns:
            Deleted repos.
        """
        repos = self._list_repo_names(repos)
        for repo in repos:
            self.adapter.delete_repo(repo)
            self.logger.info(f'Deleted repo {repo}')
        return repos

    def commit(self, repo: str, branch: Optional[str] = 'master', parent_commit: Optional[str] = None, flush: bool = False) -> 'PachydermCommit':
        """Returns a context manager for a new commit.

        The context manager automatically starts and finishes the commit.
        If an exception occurs, the commit is not finished, but deleted.

        Args:
            repo: Name of repository.
            branch: Branch in repository. When the commit is started on a branch, the previous head of the branch is
                    used as the parent of the commit. You may pass `None` in which case the new commit will have
                    no parent (unless `parent_commit` is specified) and will initially appear empty.
            parent_commit: ID of parent commit. Upon creation the new commit will appear identical to the parent commit.
                Data can safely be added to the new commit without affecting the contents of the parent commit.
            flush: If true, blocks until all jobs triggered by this commit have finished
                when the context is exited (only when leaving the `with` statement).

        Returns:
            Commit object allowing operations inside the commit.
        """
        return PachydermCommit(self, repo, branch=branch, parent_commit=parent_commit, flush=flush)

    def delete_commit(self, repo: str, commit: str) -> None:
        """Deletes a commit.

        Args:
            repo: Name of repository.
            commit: ID of commit to delete.
        """
        self.adapter.delete_commit(repo, commit)

    def create_branch(self, repo: str, commit: str, branch: str) -> None:
        """Sets a commit as a branch.

        Args:
            repo: Name of repository.
            commit: ID of commit to set as branch.
            branch: Name of the branch.
        """
        self.adapter.create_branch(repo, commit, branch)

    def delete_branch(self, repo: str, branch: str) -> None:
        """Deletes a branch, but leaves the commits intact.

        The commits can still be accessed via their commit IDs.

        Args:
            repo: Name of repository.
            branch: Name of branch to delete.
        """
        self.adapter.delete_branch(repo, branch)

    def get_file(self, repo: str, path: str, branch: Optional[str] = 'master', commit: Optional[str] = None,
                 destination: Union[str, Path, BinaryIO] = '.') -> None:
        """Retrieves a file from a repository in PFS and writes it to `destination`.

        Args:
            repo: Repository to retrieve file from.
            path: Path within repository in PFS to retrieve file from.
            branch: Branch to retrieve file from.
            commit: Commit to retrieve file from. Overrides `branch` if specified.
            destination: Local path or binary file object to write file to.
                If it is a directory the file's basename will be appended.
        """
        content = self.adapter.get_file(repo, path, branch=branch, commit=commit)
        if isinstance(destination, (str, Path)):
            if os.path.isdir(destination):
                destination = os.path.join(destination, os.path.basename(path))
            with open(destination, 'wb') as f:
                for value in content:
                    f.write(value)
        else:
            for value in content:
                destination.write(value)

    def get_file_content(self, repo: str, path: str, branch: Optional[str] = 'master', commit: Optional[str] = None,
                         encoding: Optional[str] = None) -> Union[bytes, str]:
        """Retrieves a file from a repository in PFS and returns its content.

        Args:
            repo: Repository to retrieve file from.
            path: Path within repository in PFS to retrieve file from.
            branch: Branch to retrieve file from.
            commit: Commit to retrieve file from. Overrides `branch` if specified.
            encoding: If specified, the file content will be returned as a decoded string.
                May be set to 'auto' to try to infer the encoding automatically.

        Returns:
            File contents as bytes, or as string if `encoding` is set.
        """
        with io.BytesIO() as destination:
            self.get_file(repo, path, branch=branch, commit=commit, destination=destination)
            content = destination.getvalue()
        if encoding == 'auto':
            enc = chardet.detect(content)
            encoding = enc['encoding']
            if enc['confidence'] < 0.9:
                raise ValueError('could not reliably detect encoding of file content')
        if encoding is not None:
            return content.decode(encoding)
        return content

    def get_files(self, repo: str, path: str = '/', glob: str = '**', branch: Optional[str] = 'master', commit: Optional[str] = None,
                  destination: Union[str, Path] = '.') -> None:
        """Retrieves multiple files from a repository in PFS and writes them to a local directory.

        Args:
            repo: Repository to retrieve files from.
            path: Path within repository in PFS to retrieve files from.
            glob: Glob pattern to filter files retrieved from `path`. Is interpreted relative to `path`.
            branch: Branch to retrieve files from.
            commit: Commit to retrieve files from. Overrides `branch` if specified.
            destination: Local path to write files to. Must be a directory. Will be created if it doesn't exist.
        """
        path = '/' + path.strip('/')
        glob = path + '/' + glob
        destination = Path(destination).expanduser().resolve()
        files = self.adapter.list_files(repo, branch=branch, commit=commit, glob=glob)
        files = files[files['type'] == 'file']
        for _, row in self._progress(files.iterrows(), n=len(files), unit='file'):
            local_path = destination / os.path.relpath(row['path'], path)
            os.makedirs(local_path.parent, exist_ok=True)
            self.get_file(repo, path=row['path'], commit=row['commit'], destination=local_path)
            self.logger.info(f"Downloaded '{os.path.basename(row['path'])}' to '{local_path.parent}'")

    def create_pipelines(self, pipelines: WildcardFilter = '*', pipeline_specs: Optional[List[dict]] = None,
                         recreate: bool = False, build_options: Optional[dict] = None) -> PipelineChanges:
        """Creates or recreates pipelines.

        Args:
            pipelines: Pattern to filter pipeline specs by pipeline name. Supports shell-style wildcards.
            pipeline_specs: Pipeline specifications. These are read from files
                (see property `pipeline_spec_files`) if not specified.
            recreate: Whether to delete existing pipelines before recreating them.
            build_options: Keyword arguments to pass into the Docker `build()` method when building images.
                A useful example is ``dict(nocache=True)``.

        Returns:
            Created and deleted pipeline names.
        """
        return self._create_or_update_pipelines(pipelines=pipelines, pipeline_specs=pipeline_specs,
                                                update=False, recreate=recreate, reprocess=False, build_options=build_options)

    def update_pipelines(self, pipelines: WildcardFilter = '*', pipeline_specs: Optional[List[dict]] = None,
                         recreate: bool = False, reprocess: bool = False, build_options: Optional[dict] = None) -> PipelineChanges:
        """Updates or recreates pipelines.

        Non-existing pipelines will be created.

        Args:
            pipelines: Pattern to filter pipeline specs by pipeline name. Supports shell-style wildcards.
            pipeline_specs: Pipeline specifications. These are read from files
                (see property `pipeline_spec_files`) if not specified.
            recreate: Whether to delete existing pipelines before recreating them.
            reprocess: Whether to reprocess datums with updated pipeline.
            build_options: Keyword arguments to pass into the Docker `build()` method when building images.
                A useful example is ``dict(nocache=True)``.

        Returns:
            Updated, created and deleted pipeline names.
        """
        return self._create_or_update_pipelines(pipelines=pipelines, pipeline_specs=pipeline_specs,
                                                update=True, recreate=recreate, reprocess=reprocess, build_options=build_options)

    def delete_pipelines(self, pipelines: WildcardFilter) -> List[str]:
        """Deletes existing pipelines.

        Args:
            pipelines: Pattern to filter pipelines by name. Supports shell-style wildcards.

        Returns:
            Names of deleted pipelines.
        """
        pipelines = self._list_pipeline_names(pipelines)
        for pipeline in pipelines:
            self.adapter.delete_pipeline(pipeline)
            self.logger.info(f'Deleted pipeline {pipeline}')
        return pipelines

    def start_pipelines(self, pipelines: WildcardFilter) -> List[str]:
        """Restarts stopped pipelines.

        Args:
            pipelines: Pattern to filter pipelines by name. Supports shell-style wildcards.

        Returns:
            Names of started pipelines.
        """
        pipelines = self._list_pipeline_names(pipelines)
        for pipeline in pipelines:
            self.adapter.start_pipeline(pipeline)
            self.logger.info(f'Started pipeline {pipeline}')
        return pipelines

    def stop_pipelines(self, pipelines: WildcardFilter) -> List[str]:
        """Stops running pipelines.

        Args:
            pipelines: Pattern to filter pipelines by name. Supports shell-style wildcards.

        Returns:
            Names of stopped pipelines.
        """
        pipelines = self._list_pipeline_names(pipelines)
        for pipeline in pipelines:
            self.adapter.stop_pipeline(pipeline)
            self.logger.info(f'Stopped pipeline {pipeline}')
        return pipelines

    def trigger_pipeline(self, pipeline: str, flush: bool = False) -> None:
        """Triggers a pipeline with a cron input by committing a timestamp file into its cron input repository.

        Args:
            pipeline: Name of pipeline to trigger.
            flush: If true, blocks until all triggered jobs have finished.
        """
        cron_specs = self.adapter.get_pipeline_cron_specs(pipeline)
        if len(cron_specs) == 0:
            raise PachydermClientError(f'Cannot trigger pipeline {pipeline} without cron input')
        with self.commit(repo=cron_specs[0]['repo'], flush=flush) as c:
            c.put_timestamp_file(overwrite=cron_specs[0]['overwrite'])

    def delete_job(self, job: str) -> None:
        """Deletes a job.

        Args:
            job: ID of job to delete.
        """
        self.adapter.delete_job(job)

    def read_pipeline_specs(self, pipelines: WildcardFilter = '*') -> List[dict]:
        """Read pipelines specifications from YAML or JSON files.

        The spec files are defined through the `pipeline_spec_files` property,
        which can be a list of file paths or glob patterns.

        File names are expected to be a prefix of pipeline names defined in them
        or to also match the given `pipelines` pattern.

        Args:
            pipelines: Pattern to filter pipeline specs by pipeline name. Supports shell-style wildcards.
        """
        files = self.pipeline_spec_files
        files_subset = [
            f for f in files
            if wildcard_match(os.path.basename(f), pipelines) or (isinstance(pipelines, str) and pipelines.startswith(os.path.splitext(os.path.basename(f))[0]))
        ]
        files = files_subset if len(files_subset) > 0 else files

        pipeline_specs = []
        n = len(files)
        for file in files:
            with open(file, 'r') as f:
                if file.lower().endswith('.yaml') or file.lower().endswith('.yml'):
                    file_content = yaml.safe_load(f)
                elif file.lower().endswith('.json'):
                    file_content = json.load(f)
                else:
                    n -= 1
                    continue
            if isinstance(file_content, dict):
                file_content = [file_content]
            if isinstance(file_content, list) and len(file_content) > 0:
                for i, pipeline_spec in enumerate(file_content):
                    if isinstance(pipeline_spec, dict) and all(k in pipeline_spec for k in ('pipeline', 'transform', 'input')):
                        pipeline_spec['_file'] = Path(file)
                        pipeline_specs.append(pipeline_spec)
                    else:
                        raise ValueError(f"Item {i + 1} in file '{file}' does not look like a pipeline specification")
        self.logger.debug(f'Read pipeline specifications from {n} files')

        if pipelines != '*':
            def wildcard_match_pipeline_spec(p):
                try:
                    return wildcard_match(p['pipeline'] if isinstance(p['pipeline'], str) else p['pipeline']['name'], pipelines)
                except (KeyError, TypeError):
                    return False
            pipeline_specs = [p for p in pipeline_specs if wildcard_match_pipeline_spec(p)]

        pipeline_specs = self._transform_pipeline_specs(pipeline_specs)

        if len(pipeline_specs) > 0:
            self.logger.debug(f'Pattern "{pipelines}" matched specification for {len(pipeline_specs)} pipelines')
        else:
            self.logger.warning(f'Pattern "{pipelines}" did not match any pipeline specifications')

        return pipeline_specs

    def clear_cache(self) -> None:
        self._built_images = set()
        self._image_digests = {}

    def _build_image(self, pipeline_spec: dict, build_options: Optional[dict] = None, push: bool = True) -> None:
        """Build and optionally push the Docker image specified in a pipeline specification.

        Args:
            pipeline_spec: Pipeline specification.
            build_options: Keyword arguments to pass into the Docker `build()` method.
            push: Whether to push the image to the registry after building it.
        """
        image = pipeline_spec['transform']['image']
        if image in self._built_images:
            return
        default_build_options = {'rm': True}
        pipeline_build_options = pipeline_spec['transform'].get('docker_build_options', {})
        build_options = {**default_build_options, **pipeline_build_options, **(build_options or {})}
        if 'dockerfile_path' in pipeline_spec['transform']:
            dockerfile_path = pipeline_spec['transform']['dockerfile_path']
            self.logger.info(f"Building Docker image '{image}' from '{dockerfile_path}' ...")
            build_stream = self.docker_client.api.build(path=str(dockerfile_path), tag=image, decode=True, **build_options)
        elif 'dockerfile' in pipeline_spec['transform']:
            dockerfile = io.BytesIO(pipeline_spec['transform']['dockerfile'].encode('utf-8'))
            self.logger.info(f"Building Docker image '{image}' ...")
            build_stream = self.docker_client.api.build(fileobj=dockerfile, tag=image, decode=True, **build_options)
        else:
            return
        for chunk in build_stream:
            if 'error' in chunk:
                raise DockerError(chunk['error'])
            if 'stream' in chunk:
                msg = chunk['stream'].strip()
                if msg:
                    self.logger.debug(msg)
        self._built_images.add(image)
        if push:
            registry_adapter = self._get_registry_adapter(image)
            digest = registry_adapter.push_image(image)
            self._image_digests[image] = digest
            self.logger.info(f"Successfully pushed image '{image}' to registry")
            self.logger.debug(f"Image digest is '{digest}'")

    def _transform_pipeline_specs(self, pipeline_specs: List[dict]) -> List[dict]:
        """Applies default transformations on pipeline specs.

        This includes inheritance of the `image` from previous pipelines if not specified.
        Also applies a custom transformer function on each spec if specified via property `pipeline_spec_transformer`.

        Args:
            pipeline_specs: Pipeline specifications to transform.
        """
        for pipeline in pipeline_specs:
            if callable(self.pipeline_spec_transformer):
                pipeline = self.pipeline_spec_transformer(pipeline)
            if isinstance(pipeline['pipeline'], str):
                pipeline['pipeline'] = {'name': pipeline['pipeline']}
            if 'image' not in pipeline['transform']:
                raise ValueError(f"Pipeline '{pipeline['pipeline']['name']}' is missing the `transform.image` field")
            if 'dockerfile_path' in pipeline['transform']:
                path = Path(pipeline['transform']['dockerfile_path'])
                if not path.is_absolute():
                    path = pipeline['_file'].parent / path
                if not path.is_dir() and path.name.lower() == 'dockerfile':
                    path = path.parent
                pipeline['transform']['dockerfile_path'] = path.resolve()
        return pipeline_specs

    def _get_registry_adapter(self, image: str) -> DockerRegistryAdapter:
        """Returns a registry adapter chosen based on an image string.

         - Amazon ECR if the repository contains "ecr.*.amazonaws.com"
         - Docker Registry otherwise

        Args:
            image: Image string.
        """
        if re.match(r'.+\.ecr\..+\.amazonaws\.com/.+', image):
            return self.amazon_ecr
        else:
            return self.docker_registry

    def _add_image_digest(self, image: str) -> str:
        """Adds the latest image digest to an image string.

        Args:
            image: Image string to add digest to.

        Returns:
            Updated image string (repository:tag@digest) if the latest digest was retrieved from the registry.
            Returns an unchanged image string if the image was not found in the registry or
            the image string already contained a digest.
        """
        if '@' in image:
            return image
        if image in self._image_digests:
            digest = self._image_digests[image]
        else:
            registry_adapter = self._get_registry_adapter(image)
            digest = registry_adapter.get_image_digest(image)
        if digest:
            self._image_digests[image] = digest
            return image + '@' + digest
        return image

    def _create_or_update_pipelines(self, pipelines: WildcardFilter = '*', pipeline_specs: List[dict] = None,
                                    update: bool = True, recreate: bool = False, reprocess: bool = False,
                                    build_options: Optional[dict] = None) -> PipelineChanges:
        if pipeline_specs is None:
            pipeline_specs = self.read_pipeline_specs(pipelines)
        elif pipelines != '*':
            pipeline_specs = [p for p in pipeline_specs if wildcard_match(p['pipeline']['name'], pipelines)]

        self.clear_cache()
        existing_pipelines = set(self._list_pipeline_names())
        created_pipelines: List[str] = []
        updated_pipelines: List[str] = []
        deleted_pipelines: List[str] = []

        if recreate:
            pipeline_names = [pipeline_spec['pipeline']['name'] for pipeline_spec in pipeline_specs]
            deleted_pipelines = self.delete_pipelines(pipeline_names)

        for pipeline_spec in self._progress(pipeline_specs, unit='pipeline'):
            pipeline = pipeline_spec['pipeline']['name']
            if self.build_images:
                self._build_image(pipeline_spec, build_options=build_options)
            if self.add_image_digests:
                pipeline_spec['transform']['image'] = self._add_image_digest(pipeline_spec['transform']['image'])
            if pipeline in existing_pipelines and not recreate:
                if update:
                    self.adapter.update_pipeline(pipeline_spec, reprocess=reprocess)
                    updated_pipelines.append(pipeline)
                    self.logger.info(f'Updated pipeline {pipeline}')
                else:
                    self.logger.warning(f'Pipeline {pipeline} already exists')
            else:
                self.adapter.create_pipeline(pipeline_spec)
                created_pipelines.append(pipeline)
                self.logger.info(f'Created pipeline {pipeline}')

        return PipelineChanges(created=created_pipelines, updated=updated_pipelines, deleted=deleted_pipelines)

    def _list_pipeline_names(self, match: WildcardFilter = None) -> List[str]:
        return wildcard_filter(self.adapter.list_pipeline_names(), match)

    def _list_repo_names(self, match: WildcardFilter = None) -> List[str]:
        return wildcard_filter(self.adapter.list_repo_names(), match)

    def _calc_cron_tick(self, cron_spec: str, prev: bool = True) -> Optional[pd.Timestamp]:
        if pd.isna(cron_spec) or cron_spec == '':
            return None
        now = datetime.now(self.pachd_timezone)
        cron = croniter(cron_spec, now)
        tick = cron.get_prev(datetime) if prev else cron.get_next(datetime)
        return pd.Timestamp(tick.astimezone(self.user_timezone))

    def _localize_timestamp(self, ts: pd.Series) -> pd.Series:
        return ts.dt.tz_convert(self.user_timezone)

    @classmethod
    def _progress(cls, x, n=None, **kwargs):
        del n, kwargs
        return x

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.adapter.host}:{self.adapter.port}')"


class PachydermCommit(PachydermCommitAdapter):

    """Represents a commit in Pachyderm.

    Objects of this class are typically created via :meth:`~pachypy.client.PachydermClient.commit`
    and used as a context manager, which automatically starts and finishes a commit.
    """

    def __init__(self, client: PachydermClient, repo: str, branch: Optional[str] = 'master', parent_commit: Optional[str] = None, flush: bool = False):
        self.client = client
        super().__init__(client.adapter, repo, branch=branch, parent_commit=parent_commit, flush=flush)

    def put_file(self, file: Union[str, Path, IO], path: Optional[str] = None,
                 delimiter: str = 'none', target_file_datums: int = 0, target_file_bytes: int = 0) -> None:
        """Uploads a file or the content of a file-like to the given `path` in PFS.

        Args:
            file: A local file path or a file-like object.
            path: PFS path to upload file to. Defaults to the root directory.
                If `path` ends with a slash (/), the basename of `file` will be appended.
                If `file` is a file-like object, `path` needs to be the full PFS path including filename.
            delimiter: Causes data to be broken up into separate files with `path` as a prefix.
                Possible values are 'none' (default), 'json', 'line', 'sql' and 'csv'.
            target_file_datum: Specifies the target number of datums in each written file.
                It may be lower if data does not split evenly, but will never be higher, unless the value is 0 (default).
            target_file_bytes: Specifies the target number of bytes in each written file.
                Files may have more or fewer bytes than the target.
        """
        if isinstance(file, (str, Path)):
            if path is None:
                path = '/'
            if path.endswith('/'):
                path = path + os.path.basename(file)
            with open(Path(file).expanduser().resolve(), 'rb') as f:
                self.put_file_bytes(f.read(), path, delimiter=delimiter, target_file_datums=target_file_datums, target_file_bytes=target_file_bytes)
        else:
            if path is None or path.endswith('/'):
                raise ValueError('path needs to be specified (including filename)')
            self.put_file_bytes(file.read(), path, delimiter=delimiter, target_file_datums=target_file_datums, target_file_bytes=target_file_bytes)  # type: ignore

    def put_files(self, files: FileGlob, path: str = '/', keep_structure: bool = False, base_path: Union[str, Path] = '.') -> List[str]:
        """Uploads one or multiple files defined by `files` to the given `path` in PFS.

        Args:
            files: Glob pattern(s) to files that should be uploaded.
            path: PFS path to upload files to. Must be a directory.
                Will be created if it doesn't exist. Defaults to the root directory.
            keep_structure: If true, the local directory structure is recreated in PFS.
                Use in conjunction with `base_path`.
            base_path: If `keep_structure` is true, PFS paths will be constructed
                relative to `path` and using the local file structure relative to `base_path`.
                This defaults to the current working directory.

        Returns:
            Local paths of uploaded files.
        """
        files = expand_files(files)
        for file in self.client._progress(files, unit='file'):
            if keep_structure:
                file_path = os.path.join(path, os.path.relpath(file, base_path))
            else:
                file_path = os.path.join(path, os.path.basename(file))
            if '../' in file_path:
                raise ValueError(f"'{file_path}' is not a valid path in PFS. You may want to adjust the `base_path` argument.")
            self.put_file(file, file_path)
        return files

    def put_timestamp_file(self, overwrite: bool = False) -> None:
        """Uploads a timestamp file to simulate a cron tick.

        Args:
            overwrite: Whether to overwrite existing timestamp files (True) or
                to just add a new timestamp file (False). Only applies to pachd >=1.8.6.
        """
        if self.client.pachd_version >= '1.8.6':
            if overwrite:
                self.delete_file('/time')
                self.delete_files('/????-??-??T??:??:??*')
            timestamp = datetime.utcnow().replace(microsecond=0).isoformat() + 'Z'
            self.put_file_bytes(b'', timestamp)
        else:
            self.delete_file('/time')
            timestamp = datetime.utcnow().isoformat()[:-3] + 'Z'
            self.put_file_bytes(json.dumps(timestamp).encode('utf-8'), '/time')

    def delete_files(self, glob: str) -> List[str]:
        """Deletes all files matching `glob` in the current commit.

        Args:
            glob: Glob pattern to filter files to delete.

        Returns:
            PFS paths of deleted files.
        """
        files = self._list_file_paths(glob)
        for path in files:
            self.delete_file(path)
        return files

    def __str__(self) -> str:
        return self.commit or ''

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.commit or ''}')"



import os
import json
import pytest
import pandas as pd
from unittest.mock import patch, MagicMock, PropertyMock, DEFAULT

from pachypy.client import PachydermClient, PachydermClientError


def get_mock_from_csv(file, datetime_cols=None, timedelta_cols=None, json_cols=None):
    file_path = os.path.join(os.path.dirname(__file__), 'mock', file)
    df = pd.read_csv(file_path, parse_dates=datetime_cols, converters={c: json.loads for c in json_cols or []})
    for c in datetime_cols or []:
        df[c] = df[c].dt.tz_localize('utc')
    for c in timedelta_cols or []:
        df[c] = pd.to_timedelta(df[c])
    return df


def mock_list_commits(_, repo, n=20):
    df = get_mock_from_csv('list_commits.csv', datetime_cols=['started', 'finished'], json_cols=['branches'])
    return df[df['repo'] == repo].head(n)


def mock_list_branch_heads(_, repo):
    df = mock_list_commits(_, repo, n=1)
    return {df['branches'].iloc[0][0]: df['commit'].iloc[0]}


def mock_list_files(_, repo, **kwargs):
    del kwargs
    df = get_mock_from_csv('list_files.csv', datetime_cols=['committed'], json_cols=['branches'])
    return df[df['repo'] == repo]


def mock_list_jobs(_, pipeline=None, n=20):
    df = get_mock_from_csv('list_jobs.csv', datetime_cols=['started', 'finished'], timedelta_cols=['download_time', 'process_time', 'upload_time'])
    return (df[df['pipeline'] == pipeline] if pipeline is not None else df).head(n)


def mock_list_datums(_, job):
    df = get_mock_from_csv('list_datums.csv', datetime_cols=['committed'])
    return df[df['job'] == job]


def mock_get_logs(_, pipeline=None, **kwargs):
    del kwargs
    df = get_mock_from_csv('get_logs.csv', datetime_cols=['ts'])
    return df[df['pipeline'] == pipeline] if pipeline is not None else df


def mock_inspect(_, **kwargs):
    del kwargs
    return {'key': 'value'}


def mock_get_file(_, repo, path, **kwargs):
    del repo, kwargs
    if 'random' in path:
        yield os.urandom(512)
        yield os.urandom(512)
    else:
        yield b'te'
        yield b'st'


def wrap_mock(func, empty=False):
    def wrap(*args, **kwargs):
        res = func(*args, **kwargs)
        if empty:
            if isinstance(res, pd.DataFrame):
                res = res.head(0)
            elif isinstance(res, list):
                res = []
        return res
    return wrap


def patch_adapter(empty=False):
    return patch.multiple(
        'pachypy.adapter.PachydermAdapter',
        get_version=lambda _: '1.8.6',
        list_repos=wrap_mock(lambda _: get_mock_from_csv('list_repos.csv', datetime_cols=['created']), empty),
        list_repo_names=lambda _: get_mock_from_csv('list_repos.csv')['repo'].tolist(),
        list_commits=wrap_mock(mock_list_commits, empty),
        list_branch_heads=mock_list_branch_heads,
        list_files=wrap_mock(mock_list_files, empty),
        list_pipelines=wrap_mock(lambda _: get_mock_from_csv('list_pipelines.csv', datetime_cols=['created']), empty),
        list_pipeline_names=lambda _: get_mock_from_csv('list_pipelines.csv')['pipeline'].tolist(),
        list_jobs=wrap_mock(mock_list_jobs, empty),
        list_datums=wrap_mock(mock_list_datums, empty),
        get_logs=wrap_mock(mock_get_logs, empty),
        get_file=mock_get_file,
        inspect_pipeline=mock_inspect,
        inspect_job=mock_inspect,
        inspect_datum=mock_inspect,
        create_repo=DEFAULT,
        delete_repo=DEFAULT,
        create_branch=DEFAULT,
        delete_branch=DEFAULT,
        delete_commit=DEFAULT,
        create_pipeline=DEFAULT,
        update_pipeline=DEFAULT,
        delete_pipeline=DEFAULT,
        start_pipeline=DEFAULT,
        stop_pipeline=DEFAULT,
        delete_job=DEFAULT,
        get_pipeline_cron_specs=DEFAULT
    )


def patch_commit_adapter():
    return patch.multiple(
        'pachypy.adapter.PachydermCommitAdapter',
        commit=PropertyMock,
        start=DEFAULT,
        finish=DEFAULT,
        delete=DEFAULT,
        put_file_bytes=DEFAULT,
        put_file_url=DEFAULT,
        delete_file=DEFAULT,
        _list_file_paths=DEFAULT,
    )


@pytest.fixture(scope='module')
def pipeline_spec_files_path():
    return os.path.join(os.path.dirname(__file__), 'pipelines')


@pytest.fixture(scope='function')
def client(pipeline_spec_files_path):
    return PachydermClient(
        pipeline_spec_files=os.path.join(pipeline_spec_files_path, '*.yaml'),
        add_image_digests=False,
        build_images=False
    )


@pytest.fixture(scope='module')
def docker_registry():
    import docker
    client = docker.DockerClient()
    try:
        container = client.containers.get('test_registry')
        if container.status != 'running':
            container.restart()
    except docker.errors.NotFound:
        container = client.containers.run('registry:2', name='test_registry', detach=True, ports={'5000/tcp': 5000})
    yield container
    container.stop()
    container.remove()


def test_init_registry_adapters(client: PachydermClient):
    from pachypy.registry import DockerRegistryAdapter, AmazonECRAdapter
    assert isinstance(client.docker_registry, DockerRegistryAdapter)
    assert isinstance(client.amazon_ecr, AmazonECRAdapter)


@patch_adapter()
def test_list_repos(client: PachydermClient, **mocks):
    del mocks
    df = client.list_repos()
    assert len(df) == 6
    assert df['is_tick'].sum() == 1
    assert client.list_repos('*tick').shape[0] == 1


@patch_adapter()
def test_list_commits(client: PachydermClient, **mocks):
    del mocks
    df = client.list_commits('test_x_pipeline_*')
    assert len(df) == 10
    assert df['size_bytes'].gt(0).all()
    with pytest.raises(PachydermClientError):
        assert client.list_commits('repo_that_doesnt_exist')


@patch_adapter()
def test_list_files(client: PachydermClient, **mocks):
    del mocks
    df = client.list_files('test_x_pipeline_*', files_only=True)
    assert len(df) == 4
    assert df['type'].eq('file').all()
    assert df['size_bytes'].gt(0).all()
    df = client.list_files('test_x_pipeline_3', commit='e1d7e6912d5d4a3289e9fb7c82eec6b5', files_only=False)
    assert len(df) == 7
    assert df['type'].eq('dir').any()
    df = client.list_files('test_x_pipeline_3', branch=None, commit=None, files_only=False)
    assert len(df) == 7
    with pytest.raises(PachydermClientError):
        assert client.list_files('test_x_pipeline_*', commit='e1d7e6912d5d4a3289e9fb7c82eec6b5')
    with pytest.raises(PachydermClientError):
        assert client.list_files('repo_that_doesnt_exist')


@patch_adapter()
def test_list_pipelines(client: PachydermClient, **mocks):
    del mocks
    assert len(client.list_pipelines()) == 5
    assert len(client.list_pipelines('test_x_pipeline_?')) == 5
    assert len(client.list_pipelines('test_x_pipeline_1')) == 1


@patch_adapter()
def test_list_jobs(client: PachydermClient, **mocks):
    del mocks
    df = client.list_jobs()
    assert len(df) == 8
    assert df['duration'].isna().sum() == 0
    assert df['progress'].max() <= 1
    assert len(client.list_jobs(n=5)) == 5
    assert len(client.list_jobs('test_x_pipeline_5')) == 2
    assert len(client.list_jobs('test_x_pipeline_*', n=1)) == 1
    with pytest.raises(PachydermClientError):
        assert client.list_jobs('pipeline_that_doesnt_exist')


@patch_adapter()
def test_list_datums(client: PachydermClient, **mocks):
    del mocks
    df = client.list_datums('e26ccf65131b4b3d9087cebc2f944279')
    assert len(df) == 10
    assert df['datum'].nunique() == 5
    assert df['repo'].nunique() == 2
    assert df['size_bytes'].gt(0).all()
    assert len(client.list_datums('job_id_with_no_datums')) == 0


@patch_adapter()
def test_get_logs(client: PachydermClient, **mocks):
    del mocks
    assert len(client.get_logs('test_x_pipeline_5')) == 10
    assert len(client.get_logs('test_x_pipeline_5', user_only=True)) == 7
    assert len(client.get_logs('test_x_pipeline_5', last_job_only=False)) == 20
    assert len(client.get_logs('test_x_pipeline_5', user_only=True, last_job_only=False)) == 14
    assert len(client.get_logs('test_x_pipeline_5', datum='e52eea8fb37eafbfa9d04257f41f1e403b156d63b8135cb179cf142b5f1a08d5', last_job_only=False)) == 10
    with pytest.raises(PachydermClientError):
        assert client.get_logs('pipeline_that_doesnt_exist')


@patch_adapter()
def test_create_delete_repos(client: PachydermClient, **mocks):
    del mocks
    list_repo_names = 'pachypy.adapter.PachydermAdapter.list_repo_names'
    repos = ['test_repo_1', 'test_repo_2']
    with patch(list_repo_names, MagicMock(return_value=[])):
        assert client.create_repos(repos[0]) == repos[:1]
        assert client.create_repos(repos) == repos
        assert client.delete_repos('test_repo_*') == []
    with patch(list_repo_names, MagicMock(return_value=repos[1:])):
        assert client.create_repos(repos[1]) == []
        assert client.create_repos(repos) == repos[:1]
        assert client.delete_repos('test_repo_*') == repos[1:]
    with patch(list_repo_names, MagicMock(return_value=repos)):
        assert client.create_repos(repos) == []
        assert client.delete_repos('test_repo_*') == repos
        assert client.delete_repos(repos) == repos


@patch_adapter()
def test_delete_job(client: PachydermClient, **mocks):
    job = 'a1b2c3d4'
    client.delete_job(job)
    mocks['delete_job'].assert_called_once_with(job)


@patch_commit_adapter()
def test_commit_put_files(client: PachydermClient, **mocks):
    mock_file = lambda f: os.path.join(os.path.dirname(__file__), 'mock', f)
    with client.commit('test_repo') as c:
        c.put_file(mock_file('list_commits.csv'))
        mocks['put_file_bytes'].assert_called_once()
        assert mocks['put_file_bytes'].call_args[0][1] == '/list_commits.csv'
        c.put_file(mock_file('list_files.csv'), path='/folder/')
        assert mocks['put_file_bytes'].call_args[0][1] == '/folder/list_files.csv'
        c.put_file(mock_file('list_files.csv'), path='/list_files_2.csv')
        assert mocks['put_file_bytes'].call_args[0][1] == '/list_files_2.csv'
        mocks['put_file_bytes'].reset_mock()

        with pytest.raises(ValueError):
            with open(mock_file('list_commits.csv')) as f:
                c.put_file(f)
        with open(mock_file('list_commits.csv')) as f:
            c.put_file(f, path='list_commits.csv')
        mocks['put_file_bytes'].assert_called_once()
        assert mocks['put_file_bytes'].call_args[0][1] == 'list_commits.csv'
        mocks['put_file_bytes'].reset_mock()

        c.put_files(mock_file('*.csv'))
        assert mocks['put_file_bytes'].call_count == 7


@patch_commit_adapter()
@patch_adapter()
def test_trigger_pipeline(client: PachydermClient, **mocks):
    mocks['get_pipeline_cron_specs'].return_value = []
    with pytest.raises(PachydermClientError):
        client.trigger_pipeline('pipeline_without_cron_input')

    mocks['get_pipeline_cron_specs'].return_value = [{'repo': 'pipeline_tick', 'overwrite': True}]
    client.trigger_pipeline('pipeline')
    mocks['put_file_bytes'].assert_called_once()
    mocks['put_file_bytes'].reset_mock()

    mocks['get_pipeline_cron_specs'].return_value = [{'repo': 'pipeline_tick', 'overwrite': False}]
    client.trigger_pipeline('pipeline')
    mocks['put_file_bytes'].assert_called_once()


@patch_adapter()
def test_create_delete_branch_commit(client: PachydermClient, **mocks):
    del mocks
    client.create_branch('test_repo', 'a1b2c4', 'test_branch')
    client.delete_branch('test_repo', 'test_branch')
    client.delete_commit('test_repo', 'a1b2c4')


@patch_commit_adapter()
def test_commit_delete_files(client: PachydermClient, **mocks):
    mocks['_list_file_paths'].return_value = ['/test_file_1', '/test_file_2']
    with client.commit('test_repo') as c:
        c.delete_files('test_file_*')
        assert mocks['delete_file'].call_count == 2


@patch_adapter()
def test_get_file(client: PachydermClient, tmp_path, **mocks):
    del mocks
    repo = 'test_repo'
    file = 'test_file'

    client.get_file(repo, file, destination=tmp_path)
    assert os.path.exists(tmp_path / file) and os.path.getsize(tmp_path / file) == 4

    client.get_file(repo, file, destination=tmp_path / 'renamed_test_file')
    assert os.path.exists(tmp_path / 'renamed_test_file')

    assert client.get_file_content(repo, file) == b'test'
    assert client.get_file_content(repo, file, encoding='auto') == 'test'
    assert client.get_file_content(repo, file, encoding='utf-8') == 'test'

    with pytest.raises(ValueError):
        client.get_file_content(repo, 'random_file', encoding='auto')

    client.get_files('test_x_pipeline_3', destination=tmp_path)
    files = set(tmp_path.glob('**/*.avro'))
    assert len(files) == 4
    assert tmp_path / '1/2019-03-31/1554087629/1554056117887.avro' in files


def test_read_pipeline_specs(client: PachydermClient, pipeline_spec_files_path):
    def custom_transform_pipeline_spec(pipeline_spec):
        pipeline_spec['test'] = True
        return pipeline_spec
    client.pipeline_spec_transformer = custom_transform_pipeline_spec
    assert len(client.read_pipeline_specs('test*')) == 7
    assert len(client.read_pipeline_specs('not_a_test*')) == 0
    pipeline_specs = client.read_pipeline_specs('test_a*')
    assert isinstance(pipeline_specs, list) and len(pipeline_specs) == 2
    assert pipeline_specs[0]['pipeline']['name'] == 'test_a_pipeline_1'
    assert pipeline_specs[0]['test'] is True

    client.pipeline_spec_files = os.path.join(pipeline_spec_files_path, 'test_c.json')  # type: ignore
    assert len(client.read_pipeline_specs('*')) == 1

    with pytest.raises(ValueError):
        client.pipeline_spec_files = os.path.join(pipeline_spec_files_path, 'test_d.json')  # type: ignore
        client.read_pipeline_specs('*')


def test_build_push_image(client: PachydermClient, docker_registry):
    del docker_registry
    pipeline_specs = client.read_pipeline_specs('test_e_*')
    for pipeline_spec in pipeline_specs:
        client._build_image(pipeline_spec)
        image = pipeline_spec['transform']['image']
        assert image in client._built_images
        assert image in client._image_digests
        assert client._image_digests[image].startswith('sha')
        assert client.docker_registry.get_image_digest(image) == client._image_digests[image]
    client.clear_cache()
    assert len(client._built_images) == len(client._image_digests) == 0


@patch_adapter()
def test_create_update_delete_pipelines(client: PachydermClient, **mocks):
    del mocks
    list_pipeline_names = 'pachypy.adapter.PachydermAdapter.list_pipeline_names'
    pipelines = ['test_a_pipeline_1', 'test_a_pipeline_2']
    with patch(list_pipeline_names, MagicMock(return_value=[])):
        assert client.create_pipelines('test_a*') == (pipelines, [], [])
    with patch(list_pipeline_names, MagicMock(return_value=pipelines)):
        assert client.create_pipelines('test_a*') == ([], [], [])
        assert client.update_pipelines('test_a*') == ([], pipelines, [])
        assert client.update_pipelines('test_a*', recreate=True) == (pipelines, [], pipelines)
        assert client.delete_pipelines('test_a*') == pipelines


@patch_adapter()
def test_stop_start_pipelines(client: PachydermClient, **mocks):
    del mocks
    list_pipeline_names = 'pachypy.adapter.PachydermAdapter.list_pipeline_names'
    pipelines = ['test_a_pipeline_1', 'test_a_pipeline_2']
    with patch(list_pipeline_names, MagicMock(return_value=pipelines)):
        assert client.stop_pipelines('test_a*') == pipelines
        assert client.start_pipelines('test_a*') == pipelines


def test_add_image_digest(client: PachydermClient):
    digest = 'sha1:123'
    tag = 'tag'
    with patch('pachypy.registry.DockerRegistryAdapter.get_image_digest', MagicMock(return_value=digest)) as mock:
        repo = 'user/repo'
        assert client._add_image_digest(f'{repo}:{tag}') == f'{repo}:{tag}@{digest}'
        assert client._add_image_digest(f'{repo}:{tag}') == f'{repo}:{tag}@{digest}'
        assert client._add_image_digest(f'{repo}:{tag}@sha1:000') == f'{repo}:{tag}@sha1:000'
        mock.assert_called_once_with(f'{repo}:{tag}')
    with patch('pachypy.registry.AmazonECRAdapter.get_image_digest', MagicMock(return_value=digest)) as mock:
        repo = '112233445566.dkr.ecr.xxx.amazonaws.com/repo'
        assert client._add_image_digest(f'{repo}:{tag}') == f'{repo}:{tag}@{digest}'
        mock.assert_called_once_with(f'{repo}:{tag}')
    with patch('pachypy.registry.DockerRegistryAdapter.get_image_digest', MagicMock(return_value=None)) as mock:
        for image in ['user/repo2:tag', 'user/repo2:tag@sha1:000']:
            assert client._add_image_digest(image) == image



import os
from fnmatch import fnmatch
from glob import glob
from pathlib import Path
from typing import List, Dict, Iterable, Union, Optional

import pandas as pd


WildcardFilter = Optional[Union[str, Iterable[str]]]
FileGlob = Optional[Union[str, Path, Iterable[Union[str, Path]]]]


def wildcard_filter(x: Iterable[str], pattern: WildcardFilter) -> List[str]:
    if pattern is None or pattern == '*':
        return list(x)
    else:
        return [i for i in x if wildcard_match(i, pattern)]


def wildcard_match(x: str, pattern: WildcardFilter) -> bool:
    if pattern is None or pattern == '*':
        return True
    elif isinstance(pattern, str):
        return fnmatch(x, pattern)
    else:
        return any([wildcard_match(x, m) for m in pattern])


def expand_files(files: FileGlob) -> List[str]:
    if not files:
        return []
    if isinstance(files, str) or isinstance(files, Path):
        files = [files]
    return sorted(list(set.union(*[set(glob(os.path.expanduser(f))) for f in files])))


def invert_dict(d: Dict[str, str]) -> Dict[str, List[str]]:
    inverted: Dict[str, List[str]] = {}
    for key, value in d.items():
        inverted.setdefault(value, []).append(key)
    return inverted


def to_timestamp(seconds: int, nanos: int) -> Optional[pd.Timestamp]:
    if seconds > 0:
        return pd.Timestamp(float(f'{seconds}.{nanos}'), unit='s', tz='utc')
    else:
        return None


def to_timedelta(seconds: int, nanos: int) -> pd.Timedelta:
    return pd.Timedelta(float(f'{seconds}.{nanos}'), unit='s')



import os
import time
import random
import pytest
from unittest import mock

from pachypy.adapter import PachydermAdapter, PachydermCommitAdapter, PachydermError


@pytest.fixture(scope='module')
def adapter() -> PachydermAdapter:
    return PachydermAdapter('localhost', 30650)


@pytest.fixture(scope='function')
def repo(adapter: PachydermAdapter):
    repo_name = 'test_repo_{:05x}'.format(random.randrange(16**5))
    delete_repo_if_exists(adapter, repo_name)
    adapter.create_repo(repo_name)
    yield repo_name
    delete_repo_if_exists(adapter, repo_name)


@pytest.fixture(scope='function')
def pipeline_1(adapter: PachydermAdapter):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_1'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', 'date > /pfs/out/date']
        },
        'input': {
            'cron': {
                'name': 'tick',
                'spec': '0 * * * *'
            }
        }
    })


@pytest.fixture(scope='function')
def pipeline_2(adapter: PachydermAdapter):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_2'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', 'echo "test"']
        },
        'input': {
            'cron': {
                'name': 'tick',
                'spec': '0 * * * *',
                'start': '2050-01-01T00:00:00Z'
            }
        },
        'enable_stats': True
    })


@pytest.fixture(scope='function')
def pipeline_3(adapter: PachydermAdapter):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_3'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', 'cat /pfs/*/*']
        },
        'input': {
            'union': [{
                'cron': {
                    'name': 'tick',
                    'spec': '0 * * * *'
                }
            }, {
                'pfs': {
                    'repo': 'test_pipeline_1',
                    'glob': '*'
                }
            }, {
                'pfs': {
                    'repo': 'test_pipeline_2',
                    'glob': '*'
                }
            }]
        }
    })


@pytest.fixture(scope='function')
def pipeline_4(adapter: PachydermAdapter):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_4'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', 'cat /pfs/*/*']
        },
        'input': {
            'cross': [{
                'atom': {
                    'repo': 'test_pipeline_1',
                    'branch': 'test',
                    'glob': '*'
                }
            }, {
                'atom': {
                    'repo': 'test_pipeline_2',
                    'branch': 'test',
                    'glob': '*'
                }
            }]
        }
    })


@pytest.fixture(scope='function')
def pipeline_5(adapter: PachydermAdapter, repo):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_5'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', f"sleep 5s && cat /pfs/{repo}/*" + " | wc -w | awk '{$1=$1};1' > /pfs/out/cnt"]
        },
        'input': {
            'pfs': {
                'repo': repo,
                'glob': '/'
            }
        }
    })


@pytest.fixture(scope='function')
def pipeline_6(adapter: PachydermAdapter):
    yield from pipeline(adapter, {
        'pipeline': {'name': 'test_pipeline_6'},
        'transform': {
            'image': 'alpine:latest',
            'cmd': ['/bin/sh', '-c', 'cat /pfs/*/*']
        },
        'input': {
            'union': [{
                'cron': {
                    'name': 'tick1',
                    'spec': '0 * * * *'
                }
            }, {
                'cron': {
                    'name': 'tick2',
                    'spec': '30 * * * *'
                }
            }]
        }
    })


def pipeline(adapter: PachydermAdapter, pipeline_spec):
    pipeline_name = pipeline_spec['pipeline']['name']
    delete_pipeline_if_exists(adapter, pipeline_name)
    adapter.create_pipeline(pipeline_spec)
    yield pipeline_spec
    delete_pipeline_if_exists(adapter, pipeline_name)


def delete_repo_if_exists(adapter: PachydermAdapter, repo_name):
    try:
        adapter.delete_repo(repo_name)
    except PachydermError:
        pass


def delete_pipeline_if_exists(adapter: PachydermAdapter, pipeline_name):
    try:
        adapter.delete_pipeline(pipeline_name)
    except PachydermError:
        pass


def await_pipeline_new_state(adapter: PachydermAdapter, pipeline_name, initial_state='starting', timeout=30):
    start_time = time.time()
    state = initial_state
    while state == initial_state and time.time() - start_time < timeout:
        time.sleep(1)
        pipelines = adapter.list_pipelines()
        state = pipelines.loc[pipelines.pipeline == pipeline_name, 'state'].iloc[0]
    return state


def await_job_completed_state(adapter: PachydermAdapter, pipeline_name, timeout=300):
    start_time = time.time()
    state = 'starting'
    while state in {'unknown', 'starting', 'running'} and time.time() - start_time < timeout:
        time.sleep(3)
        jobs = adapter.list_jobs(pipeline=pipeline_name, n=1)
        if len(jobs):
            state = jobs['state'].iloc[0]
    return state


def test_retry(adapter: PachydermAdapter):
    from grpc._channel import _Rendezvous
    from pachypy.adapter import retry
    state = mock.MagicMock()
    code = mock.MagicMock()
    type(code).value = mock.PropertyMock(return_value=(None, 'unavailable'))
    type(state).code = mock.PropertyMock(return_value=code)
    func = mock.MagicMock(side_effect=_Rendezvous(
        state=state,
        call=mock.MagicMock(),
        response_deserializer=mock.MagicMock(),
        deadline=mock.MagicMock(),
    ))
    adapter._max_retries = 3
    with mock.patch('pachypy.adapter.PachydermAdapter.list_repos', func):
        with mock.patch('pachypy.adapter.PachydermAdapter.check_connectivity', mock.MagicMock(return_value=True)):
            with pytest.raises(PachydermError):
                retry(adapter.list_repos)(adapter)  # type: ignore
            assert adapter._retries == adapter._max_retries
            assert func.call_count == adapter._max_retries + 1
        func.reset_mock()
        adapter._retries = 0
        with mock.patch('pachypy.adapter.PachydermAdapter.check_connectivity', mock.MagicMock(return_value=False)):
            with pytest.raises(PachydermError):
                retry(adapter.list_repos)(adapter)  # type: ignore
            assert adapter._retries == 0
            assert func.call_count == 1


def test_init(monkeypatch):
    monkeypatch.delenv('PACHD_ADDRESS', raising=False)
    adapter = PachydermAdapter()
    assert adapter.host == 'localhost' and adapter.port == 30650
    adapter = PachydermAdapter(host='test_host')
    assert adapter.host == 'test_host' and adapter.port == 30650
    with mock.patch.dict(os.environ, {'PACHD_ADDRESS': 'test_host:12345'}):
        adapter = PachydermAdapter()
        assert adapter.host == 'test_host' and adapter.port == 12345
    with mock.patch.dict(os.environ, {'PACHD_ADDRESS': 'another_test_host'}):
        adapter = PachydermAdapter()
        assert adapter.host == 'another_test_host' and adapter.port == 30650


def test_check_connectivity():
    adapter = PachydermAdapter(host='host_that_does_not_exist')
    assert adapter.check_connectivity() is False
    adapter = PachydermAdapter(host='google.com')
    assert adapter.check_connectivity() is False


def test_get_version(adapter: PachydermAdapter):
    version = adapter.get_version()
    assert len(version.split('.')) == 3


def test_create_delete_repo(adapter: PachydermAdapter):
    repo_name = 'test_repo_a1b2c'
    delete_repo_if_exists(adapter, repo_name)

    adapter.create_repo(repo_name)
    assert repo_name in set(adapter.list_repo_names())
    assert repo_name in set(adapter.list_repos().repo)

    adapter.delete_repo(repo_name)
    assert repo_name not in set(adapter.list_repo_names())


def test_create_update_delete_pipeline(adapter, pipeline_1):
    pipeline_name = pipeline_1['pipeline']['name']

    pipelines = adapter.list_pipelines()
    pipeline_names = adapter.list_pipeline_names()
    assert len(pipelines) == len(pipeline_names)
    assert pipeline_name in set(pipeline_names)
    assert pipeline_name in set(pipelines.pipeline)
    assert pipelines.loc[pipelines.pipeline == pipeline_name, 'image'].iloc[0] == pipeline_1['transform']['image']
    assert pipelines.loc[pipelines.pipeline == pipeline_name, 'cron_spec'].iloc[0] == pipeline_1['input']['cron']['spec']
    assert pipeline_name in set(adapter.list_repo_names())

    pipeline_1['transform']['image'] = 'alpine:edge'
    adapter.update_pipeline(pipeline_1)
    pipelines = adapter.list_pipelines()
    assert pipelines.loc[pipelines.pipeline == pipeline_name, 'image'].iloc[0] == 'alpine:edge'

    adapter.delete_pipeline(pipeline_name)
    assert pipeline_name not in adapter.list_pipeline_names()

    with pytest.raises(PachydermError):
        adapter.delete_pipeline('pipeline_that_does_not_exist')


def test_list_pipelines(adapter: PachydermAdapter, pipeline_1, pipeline_2, pipeline_3, pipeline_4):
    df = adapter.list_pipelines()
    assert df.shape[0] >= 4
    assert df.shape[1] == 15
    assert all([c in df.columns for c in [
        'pipeline', 'image', 'cron_spec', 'input', 'input_repos', 'output_branch',
        'parallelism_constant', 'parallelism_coefficient', 'datum_tries', 'max_queue_size',
        'jobs_running', 'jobs_success', 'jobs_failure',
        'created', 'state'
    ]])
    assert set(df.loc[df['pipeline'] == 'test_pipeline_3', 'input_repos'].iloc[0]) == {'test_pipeline_1', 'test_pipeline_2'}
    assert df.loc[df['pipeline'] == 'test_pipeline_3', 'input'].iloc[0] == '(tick ∪ test_pipeline_1:* ∪ test_pipeline_2:*)'
    assert df.loc[df['pipeline'] == 'test_pipeline_3', 'cron_spec'].iloc[0] == '0 * * * *'
    assert set(df.loc[df['pipeline'] == 'test_pipeline_4', 'input_repos'].iloc[0]) == {'test_pipeline_1', 'test_pipeline_2'}
    assert df.loc[df['pipeline'] == 'test_pipeline_4', 'input'].iloc[0] == '(test_pipeline_1/test:* ⨯ test_pipeline_2/test:*)'


def test_stop_start_pipeline(adapter: PachydermAdapter, pipeline_1):
    pipeline_name = pipeline_1['pipeline']['name']
    assert await_pipeline_new_state(adapter, pipeline_name, initial_state='starting') == 'running'

    adapter.stop_pipeline(pipeline_name)
    assert await_pipeline_new_state(adapter, pipeline_name, initial_state='running') == 'paused'

    adapter.start_pipeline(pipeline_name)
    assert await_pipeline_new_state(adapter, pipeline_name, initial_state='paused') == 'running'


def test_commit_context_manager(adapter: PachydermAdapter, repo):
    c = PachydermCommitAdapter(adapter, repo)
    assert c.commit is None and c.finished is False
    assert len(adapter.list_commits(repo)) == 0
    with c:
        pass
    assert c.commit is not None and c.finished is True
    assert len(adapter.list_commits(repo)) == 1
    assert adapter.list_branch_heads(repo)['master'] == c.commit
    with pytest.raises(PachydermError):
        with c:
            pass
    assert len(adapter.list_commits(repo)) == 1

    with pytest.raises(OSError):
        with PachydermCommitAdapter(adapter, repo) as c:
            raise OSError
    assert len(adapter.list_commits(repo)) == 1


def test_commit_put_file_bytes(adapter: PachydermAdapter, repo):
    with PachydermCommitAdapter(adapter, repo) as c:
        c.put_file_bytes('test_1', '/test_file_1')
        c.put_file_bytes('test_2', '/folder/test_file_2')
        c.put_file_bytes(b'test_3', 'test_file_3')
        c.put_file_bytes(b'test_4', 'test_file_4')
        assert set(c._list_file_paths('test*')) == {'/test_file_1', '/test_file_3', '/test_file_4'}
        c.delete_file('test_file_4')
        assert set(c._list_file_paths('test*')) == {'/test_file_1', '/test_file_3'}
        assert len(c._list_file_paths('**')) == 4
    files = adapter.list_files(repo, commit=c.commit)
    assert len(files) == 4
    assert '/test_file_1' in set(files.path)
    assert '/folder/test_file_2' in set(files.path)
    assert '/test_file_3' in set(files.path)
    assert '/test_file_4' not in set(files.path)

    with PachydermCommitAdapter(adapter, repo, branch='test') as c:
        c.put_file_bytes('test_5', '/test_file_5')
    files = adapter.list_files(repo, branch='test')
    assert len(files) == 1
    assert '/test_file_5' in set(files.path)
    assert adapter.list_branch_heads(repo)['test'] == c.commit

    commits = adapter.list_commits(repo)
    commits_branches = set(commits.branches.apply(', '.join))
    assert len(commits) == 2
    assert 'master' in commits_branches and 'test' in commits_branches
    assert c.commit in set(commits.commit)

    adapter.delete_commit(repo, c.commit)
    assert c.commit not in set(adapter.list_commits(repo).commit)

    adapter.delete_branch(repo, 'test')
    assert 'test' not in adapter.list_branch_heads(repo)


def test_commit_create_branch(adapter: PachydermAdapter, repo):
    with PachydermCommitAdapter(adapter, repo) as c:
        c.create_branch('test_branch_1')
    adapter.create_branch(repo, c.commit, 'test_branch_2')
    branch_heads = adapter.list_branch_heads(repo)
    assert 'test_branch_1' in branch_heads and 'test_branch_2' in branch_heads
    assert branch_heads['test_branch_1'] == branch_heads['master']
    assert branch_heads['test_branch_2'] == branch_heads['master']


def test_commit_put_file_url(adapter: PachydermAdapter, repo):
    with PachydermCommitAdapter(adapter, repo, branch=None) as c:
        c.put_file_url('https://raw.githubusercontent.com/itssimon/pachypy/master/tests/mock/get_logs.csv', 'get_logs.csv')
    files = adapter.list_files(repo, commit=c.commit)
    commits = adapter.list_commits(repo)
    assert len(files) == 1 and len(commits) == 1
    assert '/get_logs.csv' in set(files.path)
    assert c.commit in set(commits.commit)
    assert len(adapter.list_branch_heads(repo)) == 0


def test_commit_flush(adapter: PachydermAdapter, pipeline_5):
    pipeline = pipeline_5['pipeline']['name']
    repo = pipeline_5['input']['pfs']['repo']
    assert await_pipeline_new_state(adapter, pipeline) == 'running'
    with PachydermCommitAdapter(adapter, repo, flush=True) as c:
        c.put_file_bytes(b'a b c d e f g h i j\n', 'file1')
        c.put_file_bytes(b'k l m n o p q r s t\n', 'file2')
        t = time.time()
    assert (time.time() - t) > 5
    res = b''.join([c for c in adapter.get_file(pipeline, 'cnt')]).decode('utf-8')
    assert int(res) == 20


def test_list_commits_files(adapter: PachydermAdapter, repo):
    assert len(adapter.list_branch_heads(repo)) == 0
    assert len(adapter.list_files(repo)) == 0

    for _ in range(3):
        with PachydermCommitAdapter(adapter, repo) as c:
            c.put_file_bytes(b'test', 'test')
        with PachydermCommitAdapter(adapter, repo, branch='test') as c:
            c.put_file_bytes(b'test', 'test')

    branch_heads = adapter.list_branch_heads(repo)
    assert 'master' in branch_heads and 'test' in branch_heads

    commits = adapter.list_commits(repo, n=3)
    assert len(commits) == 3
    assert commits['repo'].iloc[0] == repo
    assert commits['size_bytes'].iloc[0] == 12
    assert commits['branches'].iloc[0] == ['test']
    assert commits['branches'].iloc[1] == ['master']
    assert commits['branches'].iloc[2] == []

    files = adapter.list_files(repo, branch='master')
    assert len(files) == 1
    assert files['repo'].iloc[0] == repo
    assert files['type'].iloc[0] == 'file'
    assert files['size_bytes'].iloc[0] == 12

    files = adapter.list_files(repo, branch='test')
    assert len(files) == 1

    with pytest.raises(ValueError):
        adapter.list_files(repo, branch=None, commit=None)


def test_list_jobs_get_logs(adapter: PachydermAdapter, pipeline_2):
    pipeline_name = pipeline_2['pipeline']['name']
    tick_repo_name = pipeline_name + '_' + pipeline_2['input']['cron']['name']
    assert await_pipeline_new_state(adapter, pipeline_name, initial_state='starting') == 'running'

    with PachydermCommitAdapter(adapter, tick_repo_name) as c:
        c.put_file_bytes(b'0', 'time')
    assert await_job_completed_state(adapter, pipeline_name) == 'success'

    pipeline_info = adapter.inspect_pipeline(pipeline_name)
    assert pipeline_info['pipeline']['name'] == pipeline_name
    assert pipeline_info['state'] == 'running'
    assert pipeline_info['jobCounts']['success'] > 0
    assert pipeline_info['version'] == 1
    assert pipeline_info['createdAt'].tzinfo is not None

    jobs = adapter.list_jobs(pipeline=pipeline_name)
    assert len(jobs) == 1
    job = jobs['job'].iloc[0]
    assert (jobs['finished'] - jobs['started']).dt.total_seconds().round().iloc[0] > -10
    assert jobs['data_processed'].iloc[0] == jobs['data_total'].iloc[0] == 1
    assert jobs['data_skipped'].iloc[0] == 0

    job_info = adapter.inspect_job(job)
    assert job_info['job']['id'] == job
    assert job_info['pipeline']['name'] == pipeline_name
    assert job_info['state'] == 'success'
    assert job_info['dataTotal'] > 0

    datums = adapter.list_datums(job=job)
    assert len(datums) == 1
    assert datums['job'].iloc[0] == job
    assert datums['repo'].iloc[0] == tick_repo_name
    assert datums['size_bytes'].iloc[0] > 0

    datum_info = adapter.inspect_datum(job=job, datum=datums['datum'].iloc[0])
    assert datum_info['datum']['job']['id'] == job
    assert datum_info['state'] == 'success'
    assert datum_info['stats']['processTime'] > 0

    logs = adapter.get_logs(pipeline=pipeline_name)
    logs = logs[logs['user']]
    assert logs.shape == (1, 7)
    assert logs['message'].iloc[0] == 'test'

    adapter.delete_job(job)
    assert len(adapter.list_jobs(pipeline=pipeline_name)) == 0


def test_get_file(adapter: PachydermAdapter, repo):
    with PachydermCommitAdapter(adapter, repo) as c:
        c.put_file_bytes(b'123', '/test_file_1')
    with PachydermCommitAdapter(adapter, repo, branch='test') as c:
        c.put_file_bytes(b'321', '/test_file_2')
    assert next(adapter.get_file(repo, '/test_file_1')) == b'123'
    assert next(adapter.get_file(repo, '/test_file_2', branch='test')) == b'321'
    assert next(adapter.get_file(repo, '/test_file_2', commit=c.commit)) == b'321'


def test_pipeline_input_cron_specs(adapter: PachydermAdapter, pipeline_5, pipeline_6):
    assert len(adapter.get_pipeline_cron_specs(pipeline_5['pipeline']['name'])) == 0
    assert len(adapter.get_pipeline_cron_specs(pipeline_6['pipeline']['name'])) == 2



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pachypy',
    version='0.1.3',
    author='Simon Gurcke',
    description='Python client library for Pachyderm',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/itssimon/pachypy',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=[
        'python_pachyderm>=1.8.6',
        'docker>=3.7.2',
        'boto3>=1.9.119',
        'pandas>=0.24.2',
        'pyyaml>=3.13',
        'croniter>=0.3.29',
        'pytz>=2018.9',
        'tzlocal>=1.5.1',
        'termcolor>=1.1.0',
        'ipython>=7.4.0',
        'tqdm>=4.31.1',
    ],
    extras_require={
        'docs':  [
            'sphinx>=2.0.0',
            'sphinx_rtd_theme>=0.4.3',
            'sphinx_autodoc_typehints>=1.6.0',
        ],
        'test': [
            'pytest>=4.3.1',
            'pytest-cov>=2.6.1',
        ]
    }
)
